from collections import defaultdict
import random
from copy import  deepcopy
import operator
from .line import *
from .relevance_functions_extended import *

# NOTE: when devising euclidean_point_distance measures
'''
reference count
vector -> ed vector -> (ed in bounds):float -> bool(float)

reference existence
vector -> ed vector -> (ed in bounds):bool
'''
lambda_floatin = lambda x,b: x >= min(b) and x <= max(b)

lambda_pointinbounds = lambda p,b: point_in_bounds_(b,p)

def lambda_countpointsinbounds(p,b):
    if b.shape[0] == 1:
        mask = np.logical_and(p >= b[0,0],p <= b[0,1])
    else:
        mask = np.logical_and(p >= b[:,0],p <= b[:,1])
    q = len(np.where(mask == True)[0])
    return q

def lambda_ratiopointsinbounds(p,b):
    x = lambda_countpointsinbounds(p,b)
    return zero_div(x,len(p),np.inf)

def random_select_k_unique_from_sequence(s, k):
    s = list(s)
    assert len(s) >= k
    random.shuffle(s)
    return s[:k]


################################### TODO: delete these functions after refactor

# TODO:
def relevance_zoom_func_1(referencePoint,boundsDistance,activationThreshold):
    # TODO: check for 0-case
    assert boundsDistance >= 0.0, "invalid bounds distance {}".format(boundsDistance)
    assert activationThreshold >= 0.0 and activationThreshold <= 1.0, ""
    return lambda p: euclidean_point_distance(p, referencePoint) <= activationThreshold * boundsDistance

def relevance_zoom_func_2(referencePoint, modulo, activationThreshold):
    rp = np.array(referencePoint, dtype = "int")
    lim = referencePoint.shape[0]
    def p(x):
        q = (rp - np.array(x,dtype="int")) % 2
        return lambda p: True if len(q == 1.0) >= lim else False
    return p


## TODO: delete?
def relevance_func_2(modulo, moduloPercentileRange):
    '''
    a sample relevance function to help demonstrate the work of CenterResplat.

    :return: function(vector) -> True if all values in (point % modulo) fall within moduloPercentileRange
    :rtype: function<vector>
    '''
    assert modulo >= 0.0, "invalid modulo"
    assert is_valid_point(moduloPercentileRange)
    assert min(moduloPercentileRange) >= 0.0 and max(moduloPercentileRange) <= 1.0
    assert moduloPercentileRange[0] <= moduloPercentileRange[1]

    minumum,maximum = moduloPercentileRange[0] * modulo,moduloPercentileRange[1] * modulo

    def f(p):
        p_ = p % modulo
        return np.all(p_ >= minumum) and np.all(p_ <= maximum)

    return f

def vector_modulo_function_with_addon(modulo, addOn):
    """
    boolean function, addon determines
    """
    def x(v):
        q = np.array(v,dtype="int")
        v_ = q % modulo
        return addOn(v_)
    return x

##### add-on functions : vector -> bool

"""
"""
def vf_vector_reference(vr, pw):
    def x(v):
        return pw(vr,v)
    return x

def subvector_iselector(indices):
    def a(v):
        return v[indices]
    return a


def m(v,addOn,iov,outputType):
    """
    m is index|value selector function for arg. to func<addOn>
    """
    assert iov in [0,1,2]

    x = []
    for t in enumerate(v):
        qi = t[iov] if iov != 2 else t
        print("QI: ",qi)
        if addOn(qi):
            if outputType == 1: q = t[1]
            else: q = t
            x.append(q)
    return np.array(x)

def subvector_selector(addOn, inputType =1, outputType = 1):
    """
    outputs elements by addOn

    :param addOn: function that outputs either singleton or pair
    :type addOn: function
    """
    assert inputType in [0,1,2], "invalid input"
    assert outputType in [1,2], "invalid output"

    def m_(v):
        return m(v,addOn,inputType,outputType)

    return m_

## $
def addon_singleton__bool__criteria_range_each(rangeReq):
    '''
    :param rangeReq: proper bounds vector
    :type rangeReq: proper bounds vector
    :return:
    :rtype: function<vector>->bool
    '''

    assert is_proper_bounds_vector(rangeReq), "invalid ranges"
    # rangeReq := length 1 or vr.shape[0]
    if rangeReq.shape == (1,2):
        q = rangeReq[0]
        p = lambda v: np.all(v >= q[0]) and np.all(v <= q[1])
    else:
        p = lambda v: point_in_bounds(rangeReq,v)
    return p

def addon_singleton__bool__criteria_distance_from_reference(rf, dm, dt,cf):
    """
    :param rf: reference value
    :type rf: ?
    :param dm: distance measure between (rf,v)
    :type dm: func<v1::vector,v2::vector>->float
    :param dt: distance threshold
    :type dt: float
    :param cf: comparator function on (dist,dt)
    :type cf: function<?>->?
    :return: function on `cf`
    :type: function<cf>->?
    """
    return lambda v: cf(dm(rf,v),dt)

class RCInst:
    """
    class that acts as a node-like structure and a function, node is designed to be used w/ the capabilities:
    
    (1) referential data (from outside of chain)
    
    (2) (standard operator,operand)

    This class is a function on a vector `v` that works by one of the four paths:
    
    (1) deciding path with reference : `cf(dm(rf,v),dt)`.
    
    (2) deciding path w/o reference: `cf(v,dt)`.
    
    (3) non-deciding path w/ reference: `cf(dm(rf,v))`.
    
    (4) non-deciding path w/o reference: `cf(v)`.

    Deciding paths require a none-null threshold value `dt`.

    In addition to these capabilities, :class:`RCInst` can update its variables.  

    :attribute rf: reference value, as argument to dm(v,rf)
    :attribute dm: function f(rf,v)
    :attribute cf: function f(v,*dt), examples include  operator.lt, operator.le, lambda_floatin, np.cross
    :attribute dt: value, use in the case of decision
    """

    def __init__(self):
        self.rf = None
        self.dm = None
        self.cf = None
        self.dt = None
        #: key is one of `rf`, `dt`; value is update function for that variable
        self.updateFunc = {}
        #: list of values used in accordance with update function in `self.updateFunc` by indices specified in `self.updatePath`
        self.updateInfo = None
        #: key is one of `rf`, `dt`; value is list of indices corresponding to values in `self.updateInfo`
        self.updatePath = {}

    def __str__(self):
        s = "* RCInst contents\n\t"
        s += str(self.rf) + "\n\t"
        s += str(self.dm) + "\n\t"
        s += str(self.cf) + "\n\t"
        s += str(self.dt) + "\n\t"
        return s

    def inst_update_var(self):
        ''' updates each key `k` in `updatePath` by the following:
        for `k`, retrieve all relevant variables `q` from `updateInfo`
        by the indices specified by `updatePath`[k] and then 
        then uses the update function `updateFunc`[k] on `q` to 
        produce the updated value for variable `k`. 
        '''

        for k,v in self.updatePath.items():

            # fetch function args
            q = []
            for v_ in v:
                q.append(self.updateInfo[v_])

            f = self.updateFunc[k]
            x = f(*tuple(q))
            self.update_var(k,x)

    def load_update_info(self,updateInfo):
        self.updateInfo = updateInfo

    def update_var(self,k,v):
        '''
        updates one of the class attributes `rf`, `dm`, `cf`,or `dt`
        with value `v`.

        :param k: attribute name
        :type k: str
        :param v: new update value
        :type v: ?
        '''

        if k == "rf":
            self.rf = v
        elif k == "dm":
            self.dm = v
        elif k == "cf":
            self.cf = v
        elif k == "dt":
            self.dt = v
        else:
            raise ValueError("invalid key _{}_".format(k))

    def mod_cf(self,dcf):
        self.cf = dcf(self.cf)
        return

    def set_reference_data(self,rf,dm):
        
        self.load_var_ref(rf)
        self.load_var_dm(dm)
        return

    ############# some functions

    def branch_at(self,n,i):
        return -1

    def load_var_ref(self,rf):
        self.rf = rf

    def load_var_cf(self,cf):
        self.cf = cf

    def load_var_dt(self,dt):
        '''
        loads a new deciding threshold value `dt`; used as `cf(v,dt)`
        '''

        self.dt = dt

    def path_type(self):
        if type(self.rf) != type(None):
            return "ref"
        return "dir"

    # TODO: class does not know.
    def output_type():
        return -1

    def load_var_dm(self,dm):
        """
        :param dm: function on (rf,v)
        """
        self.dm = dm

    def load_path(self):
        '''
        instantiates the main function `self.f` by one of the four specifications
        in the description for this class
        '''
        
        # deciding path
        if type(self.dt) != type(None):
            # for output type bool|float
            if self.path_type() == "ref":
                # calculates distance from .reference
                self.f = lambda v: self.cf(self.dm(self.rf,v),self.dt)
            else:
                self.f = lambda v: self.cf(v, self.dt)
        # non-deciding path
        else:
            if self.path_type() == "ref":
                # calculates distance from .reference
                self.f = lambda v: self.cf(self.dm(self.rf,v))
            else:
                self.f = lambda v: self.cf(v)
        return deepcopy(self.f)

class RChainHead:
    """
    :class:`RChainHead` is a sequence of functions with each element a node-like structure,represented by `RCInst`,
    that acts as a 'super' function. 
    
    Each node of this class is modifiable by update.
    ------------------------------------------------

    To make an :class:`RCInst` at index `i` of the :class:`RChainHead` instance updatable at a variable `v`,
    set its update function in the :class:`RCInst` dictionary class variable `updateFunc` for `v`, set the 
    variable path index vector `updatePath` for the same variable.

    Then update the `updatePath` variable of the :class:`RChainHead` instance with
    key `i` (denoting the node index of the :class:`RCInst` instance) with a vector of
    indices corresponding to the `variablesList` parameter in the 
    :class:`RChainHead`.load_update_vars function.

    Example:
                ```
                rch.s[0].updateFunc = {'rf': update_dt_function}
                rch.s[0].updatePath = {'rf':[0]}
                rch.updatePath = {0: [0]}

                rch.load_update_vars([one,two,three,four])
                ```
        * :class:`RChainHead` instance loads list of variables, and the variable
          `one` is loaded into the :class:`RCInst` at index 0, and `one` is used to
          update the reference value `rf` of the :class:`RCInst` instance. 


    :attribute s: the chain of class<RCInst> instances.
    :type s: list(`RCInst`)
    :attribute vpath: the sequence of transformation values that a value `v` goes through at each node in `s`. 
    :type vpath: list(values) 
    :param updatePath: a dictionary of the key-value form `node index -> sequence of indices in arg<varList>`.
    :type updatePath: dict
    :param verbose: for print-display at each class<RCInst> transformation.
    :type verbose: bool
    """
    
    def __init__(self,verbose = False):
        self.s = []
        self.vpath = []
        self.updatePath = {}
        self.verbose = verbose

    def update_rch(self):
        '''
        Updates variables of each node class<RCInst>.

        - example:
        class<ResplattingSearchSpaceIterator> calls this method on its class<RChainHead> instance
        after iterating through one pertinent bound. 
        '''

        for s_ in self.s:
            s_.inst_update_var()

    def load_update_path(self,up):
        '''
        loads a `self.updatePath`

        :param updatePath: update information (see )
        '''

        self.updatePath = up
        return

    def load_update_vars(self,varList):
        '''
        :param varList: tuple of update-variables
        '''

        for k,v in self.updatePath.items():
            vl = []
            for (i,v2) in enumerate(varList):
                if i in set(v):
                    vl.append(v2)
            self.s[k].load_update_info(vl)

    def link_rch(self,rch,linkerFunc, prev = False):
        if prev:
            return linkerFunc(self,rch)
        return linkerFunc(self,rch)

    def vpath_subset(self,si):
        return [x for (i2,x) in enumerate(self.vpath) if i2 in si]

    def load_cf_(self, rci,cfq):
        '''
        '''

        if type(cfq) == type(()):
            xs = tuple(self.vpath_subset(cfq[1]))
            cf = cfq[0](*xs)

            # below method
            rci.load_var_cf(cf)
        else:
            rci.load_var_cf(cfq)

    def make_node(self,kwargz):
        '''
        instantiates an `RCInst` node using the argument
        sequence `kwargz`.
        Note: selectorIndices refer to values in `vpath`.

        :param kwargz: If index 0 is `r` (node uses reference values), then
                format is (`r`,rf,dm|(dm,selector indices),cf|(cf,selectorIndices),?dt?).
                If index 0 is `nr` (node does use reference values), then
                format is (`nr`,cf|(cf,selectorIndices),?dt?).
                
                Please see the description for `RCInst` for details on these values.
        :type kwargz: iterable
        '''

        assert kwargz[0] in ["r","r+","nr"]

        rci = RCInst()
        if kwargz[0] == "r":
            assert len(kwargz) in [4,5], "invalid length for kwargs"
            rci.set_reference_data(kwargz[1],kwargz[2])
            self.load_cf_(rci,kwargz[3])
            try: rci.load_var_dt(kwargz[4])
            except: pass

        elif kwargz[0] == "nr":
            assert len(kwargz) in [2,3], "invalid length for kwargs"
            self.load_cf_(rci,kwargz[1])
            try: rci.load_var_dt(kwargz[2])
            except: pass

        elif kwargz[0] == "r+":
            return -1

        rci.load_path()
        return rci

    def add_node_at(self, kwargz, index = -1):
        '''

        '''

        assert index >= -1, "invalid index"
        n = self.make_node(kwargz)
        if index == -1:
            self.s.append(n)
        else:
            self.s.insert(index,n)
        return -1

    def apply(self,v):
        '''
        main function of RChainHead; applies the composite function (full function path) onto v
        
        :param v: argument into chain function
        :type v: ?
        '''

        i = 0
        v_ = np.copy(v)
        self.vpath = [v_]
        
        if self.verbose:
            print("-- applying function on {}".format(v))

        while i < len(self.s):
            q = self.s[i]
            v_ = q.f(v_)
            if self.verbose:
                print("\t-- new value: {}".format(v_))

            self.vpath.append(v_)
            i += 1
        return v_

    def cross_check(self):
        return -1

    def merge(self):
        return -1

    def __next__(self):
        return -1

####----------------------------------------------------------

###### START: helper functions for next section

def boolies(v_):
    '''
    :return: is boolean vector all true?
    :rtype: bool
    '''

    return v_ == True

def column_selector(columns, flatten = False):

    def p(v):
        q = v[:,columns]
        if flatten: return q.flatten()
        return q
    return p

def vector_index_selector(indices):
    def p(v):
        return v[indices.astype('int')]
    return p

def vector_index_inverse_selector(v):
    def p(indices):
        return v[indices.astype('int')]
    return p


###### END: helper functions for next section

###### START: functions used for relevance zoom

'''
'''
def RCHF__point_in_bounds(b):
    kwargs = ['nr', lambda_pointinbounds, b]
    # f : v -> v::bool
    rc = RChainHead()
    rc.add_node_at(kwargs)
    return rc.apply

def hops_to_default_noise_range(h):
    return np.array([[(h ** -1) / 2.7, (h ** -1) / 2.3]])

#################################### start : ostracio && deletira

def hops_to_coverage_points__standard(k,h):
    """

    arguments:
    - k :=
    - h := hop value of SSI

    return:
    - matrix, dim (m,k), m the required number of points
    """

    z = np.zeros((k,))
    o = np.ones((k,))

    #
    partition = n_partition_for_bound(np.array([z,o]).T,h)

    # case: odd
    if h % 2:
        x = [i for i in range(h + 1) if i % 2]
    # case: even
    else:
        x = [i for i in range(h + 1) if not i % 2]
    return partition[x]

# TODO: test this.
def hops_to_coverage_points_in_bounds(parentBounds,bounds,h):

    k = bounds.shape[0]
    cp = hops_to_coverage_points__standard(k,h)

    if is_proper_bounds_vector(bounds):
        s = [point_on_bounds_by_ratio_vector(bounds,c) for c in cp]
    else:
        s = [point_on_improper_bounds_by_ratio_vector(\
            parentBounds,bounds,c) for c in cp]
    return np.array(s)

def coverage_ratio_to_distance(boundsEDistance, h,cr):
    '''
    given a coverage ratio `cr`, calculates its corresponding distance to the
    euclidean distance of a bounds `boundsEDistance`. 
    
    :param boundsEDistance: euclidean distance pertaining to a bounds
    :type boundsEDistance: float
    :param h: hop
    :type h: int|float
    :param cr: 0 <= x <= 1
    :type cr: float
    '''

    assert h >= 1.0, "invalid h"
    assert cr >= 0.0 and cr <= 1.0, "invalid cr"

    total = boundsEDistance / h
    return total * cr

#################################### end : ostracio && deletira


'''
'''
def RCHF__point_in_bounds_subvector_selector(b):
    def qf(xi):
        return operator.le(xi[1],b[xi[0],1]) and operator.ge(xi[1],b[xi[0],0])

    q2 = subvector_selector(qf,inputType = 2,outputType = 1)
    kwargs = ['nr', q2]
    rc = RChainHead()
    rc.add_node_at(kwargs)
    return rc.apply

from .poly_struct import *

def RCHF__ISPoly(x:'float',largs):
    """constructs a :class:`RChainHead` function that has as its first function
    a polynomial function represented by :class:`ISPoly` over float `x`. The
    remaining nodes of the :class:`RChainHead` instance are constructed by the
    sequence of arguments `largs` to produce a :class:`RChainHead` with 1+ nodes.
    
    :param x: value that single-variable polynomial applies over
    :type x: float
    :param largs: sequence of `kwargs` used by `RChainHead.make_node`
    """

    rc = RChainHead()

    isp = ISPoly(x)

    def qf(v):
        return isp.apply(v)

    kwargs = ['nr',qf]
    rc.add_node_at(kwargs)

    for a in largs:
        rc.add_node_at(a)
    return rc.apply

def RCHF___in_bounds(bounds0):
    '''
    non-referential class<RChainHead> instance that outputs the func for in bounds
    '''

    kwargs = ['nr', lambda_pointinbounds, bounds0]
    # f : v -> v::bool
    rc = RChainHead()
    rc.add_node_at(kwargs)

    # f : filter out True | False
    subvectorSelector = boolies
    ss = subvector_selector(subvectorSelector,2)
    kwargs = ['nr',ss]
    rc.add_node_at(kwargs)

    # f : get indices
    ss = column_selector([0],True)
    kwargs = ['nr',ss]
    rc.add_node_at(kwargs)

    # f : apply indices on reference
    kwargs = ['nr',(vector_index_inverse_selector,[0])]
    rc.add_node_at(kwargs)

    return rc.apply

def ffilter(v,f):
    '''
    filters out elements of vector `v` by boolean function `f` into
    two lists `TRUE` and `FALSE`.
    '''
    t,l = [],[]
    for v_ in v:
        if f(v_): t.append(v_)
        else: l.append(v_)

    return t,l

def rpmem_func(rf,rOp):
    '''
    using reference rf,
    odd_multiply
    even

    [0] multiplier of even indices
    [1] multiplier of odd indices

    recent memory
    * + - => -
    * - - => -
    * - + => +
    * + + => +

    past memory
    * + - => +
    * - - => -
    * - + => -
    * + + => +
    '''

    rfo0,rfo1 = ffilter(rf,lambda i: i % 2)
    r1 = np.product(rfo1) if rOp else np.product(rfo0)

    def p(v):
        v0,v1 = ffilter(v,lambda i: i % 2)
        r2 = np.product(v1) if rOp else np.product(v0)
        return int(r2) % 2

    # try swapping them

    return p


def is_valid_pm(pm):
    """
    is permutation map valid?
    """
    assert pm.shape[1] == 2, "invalid pm shape"
    tf1 = len(np.unique(pm[:,0])) == pm.shape[0]
    tf2 = len(np.unique(pm[:,1])) == pm.shape[0]
    return tf1 and tf2

def is_proper_pm(pm):
    """
    is permutation map proper?

    "proper" := valid and all valids in range [0,n-1]
    """
    s = is_valid_pm(pm)
    if not s: return s
    s1 = min(pm[:,0]) == 0 and max(pm[:,0])  == pm.shape[0]
    s2 = min(pm[:,1]) == 0 and max(pm[:,1])  == pm.shape[0]
    return s1 and s2

# a version of rp mem that uses a permutation map
# TODO
def rpmem_func__pm(rfs,pm):
    """
    the original func<rpmem_func> operates on the binary choice
    """
    assert is_proper_pm(pm), "[0] invalid permutation map"
    assert pm.shape[0] == len(rfs) + 1,"[1] invalid permutation map"
    return -1

def is_valid_subset_sequence(s,n):
    q = []
    for s_ in s:
        q.extend(list(s))

    tf0 = len(q)
    q = np.unique(q)
    tf1 = len(q)
    if tf0 != tf1: return False

    m0,m1 = min(q),max(q)
    return m0 == 0 and m1 == n - 1

###### END: functions used for relevance zoom


###### a sample RCH w/ update functionality

def sample_rch_1_with_update(parentBounds, bounds, h, coverageRatio):
    '''
vector -> e.d. vector -> (ed in bounds):bool
    '''
    def dm(rp,v):
        #print("LENGO: ",len(rp))
        return np.array([euclidean_point_distance(v,rp_) for rp_ in rp])

    def cf(ds,dt_):
        return np.any(ds <= dt_)

    def update_dt_function(parentBounds,bounds,h,coverageRatio):
        return (euclidean_point_distance_of_bounds(parentBounds,bounds) / h)\
                    * coverageRatio

    rch = RChainHead()
    # add the node
    rf = hops_to_coverage_points_in_bounds(parentBounds,bounds,h)
    dm = dm
    cf = cf
    ed = euclidean_point_distance_of_bounds(parentBounds,bounds)

    # WRONG
    dt = coverage_ratio_to_distance(ed,float(h),coverageRatio)

    kwargs = ['r',rf,dm,cf,dt]
    rch.add_node_at(kwargs)

    # add update functionality
    rch.s[0].updateFunc = {'rf': hops_to_coverage_points_in_bounds,\
            'dt': update_dt_function}
    rch.s[0].updatePath = {'rf': [0,1,2],'dt':[0,1,2,3]}
    rch.updatePath = {0: [0,1,2,3]}
    return rch

def sample_rch_2_with_update(parentBounds, bounds):
    """
    one node; updates `dt` every batch with new `bounds` parameter.

    For each point, outputs true if it is in 0-20 percentile or 80-100 percentile
    of area in `bounds`. 
    """

    def activation_range(parentBounds,bounds):
        v = np.ones((parentBounds.shape[0],)) * 0.2
        b1s = np.copy(bounds[:,0])
        b1e = point_on_improper_bounds_by_ratio_vector(parentBounds,bounds,v)

        v = np.ones((parentBounds.shape[0],)) * 0.8
        b2s = point_on_improper_bounds_by_ratio_vector(parentBounds,bounds,v)
        b2e = np.copy(bounds[:,1])

        B1 = np.vstack((b1s,b1e)).T
        B2 = np.vstack((b2s,b2e)).T
        return B1,B2

    def update_dt_function(parentBounds,bounds):
        b1,b2 = activation_range(parentBounds,bounds)
        return (np.copy(parentBounds),b1,b2)

    """
    r := parentBounds,b1,b2
    """
    def cf(p,r):
        if point_in_improper_bounds(r[0],r[1],p):
            return True
        if point_in_improper_bounds(r[0],r[2],p):
            return True
        return False

    rch = RChainHead()
    dt = update_dt_function(parentBounds,bounds)
    kwargs = ['nr',cf,dt]
    rch.add_node_at(kwargs)

    # add update functionality
    rch.s[0].updateFunc = {'dt': update_dt_function}
    rch.s[0].updatePath = {'dt':[0,1]}
    rch.updatePath = {0: [0,1]}

    return rch

def sample_rch_3_with_update(modulo,updateAdder,updateModulo,targetValueRange):
    '''
    class<RChainHead> instance has one node that updates reference values and deciding threshold
    using function<vector_summation_modulo>. 

    use simple bounds and selector indices.

    NOTE: class<RChainHead> is not compatible with class<ResplattingSearchSpaceIterator> and is used
    to test accuracy and capability of class<RChainHead>.  
    '''

    assert targetValueRange[0] <= targetValueRange[1], "invalid targetValueRange"
    
    
    def cf(x,t):
        return x >= t[0] and x <= t[1]
    
    """
    def cf(x): 
        return x
    """
    rch = RChainHead(False)

    # declare the node
    kwargs = ['r',modulo,vector_summation_modulo,cf,targetValueRange]
    #kwargs = ['r',modulo,vector_summation_modulo,cf]

    rch.add_node_at(kwargs)

    # add the update function for argument `modulo`; 
    # update function does not require additional variable
    update_dt_function = update_modulo_function(updateAdder,updateModulo)
    rch.s[0].updateFunc = {'rf': update_dt_function}
    rch.s[0].updatePath = {'rf':[0]}
    rch.updatePath = {0: [0]}
    rch.load_update_vars([modulo]) 

    return rch

def sample_rch_constant_mapper(c):
    '''
    one node; maps any vector to <c> vector. 
    '''

    def l(v):
        x = np.ones((len(v),))
        return x * c

    kwargs = ['nr',l]
    rc = RChainHead()
    rc.add_node_at(kwargs)
    return rc

def sample_rch_blind_accept():
    '''
    one node; deciding + non-referential, outputs true for all vectors.
    '''

    def accept(v,i):
        return type(v) == type(i)

    rc = RChainHead()
    idv = np.ones(10)
    kwargz = ["nr",accept,idv]
    rc.add_node_at(kwargz)
    return rc

def idv(v):
    return v

def sample_rch_hop_counts_scheme_type_1(parentBound,bound,hop,cf=idv):
    '''
    one node; requires a parent bound,bound,and hop as update values
    for argument `p` (vector point).
    '''

    def f(r,v):
        return bounded_point_to_hop_counts(r[0],r[1],v,r[2])
    
    def idrf(x1,x2,x3):
        return (x1,x2,x3)

    rch = RChainHead()
    kwargs = ['r',(parentBound,bound,hop),f,cf]
    rch.add_node_at(kwargs)

    rch.s[0].updateFunc = {'rf': idrf} 
    rch.s[0].updatePath = {'rf': [0,1,2]}
    rch.updatePath = {0: [0,1,2]}
    return rch

def index_map_summation_scheme_type_1(m2,m):
    '''
    calculates a vector of indices for v based on a summation scheme.
    '''
    assert m > m2, "modulo has to be greater than v"

    def indexia(v):
        vx_ = []
        c = 1
        l = len(v) + 1
        s = np.sum(v)
        s = int(s * 10 ** decimal_places_of_float(s))
        while c < l:
            s2 = (s * c) % m
            if s2 >= l - 1: break
            vx_.append(s2)
            c += 1
        return np.array(vx_)
    return indexia

def subvector_selector_summation_scheme_type_1(m2,m):
    qf = index_map_summation_scheme_type_1(m2,m)

    def f(v):
        iv = qf(deepcopy(v))
        return np.array([v[i] for i in iv])
    return f


def sample_rch_hop_counts_scheme_type_2(parentBound,bound,hop,m2,cf=idv):
    '''
    one node; requires a parent bound,bound,and hop as update values
    for argument `p` (vector point).
    '''

    qf = subvector_selector_summation_scheme_type_1(bound.shape[0],m2)
    
    def f(r,v):
        #return qf(bounded_point_to_hop_counts(r[0],r[1],v,r[2]))
        return qf(v) 
    
    def idrf(x1,x2,x3):
        return (x1,x2,x3)

    rch = RChainHead()
    kwargs = ['r',(parentBound,bound,hop),f,cf]
    rch.add_node_at(kwargs)

    rch.s[0].updateFunc = {'rf': idrf} 
    rch.s[0].updatePath = {'rf': [0,1,2]}
    rch.updatePath = {0: [0,1,2]}
    return rch


def binary_labelling_scheme_1(v,m = 7,m2 = 4):
    '''
    outputs a vector v' of equal length to v.

    '''
    if len(v) == 0: return np.array([])
    f = decimal_order_of_vector(v) 
    v = [int(v_ * 10 ** f) for v_ in v]
    return np.array([1. if v_ % m >= m2 else 0. for v_ in v])

def sample_rch_binary_labelling_scheme_1():
    '''
    uses default parameters `m` and `m2` of method<binary_labelling_scheme_1>
    '''
    rch = RChainHead()
    kwargs = ['nr',binary_labelling_scheme_1]
    rch.add_node_at(kwargs)
    return rch


from morebs2.ball_comp import *
from morebs2.message_streamer import *
from morebs2.ball_comp_test_cases import *

def test__BallComp__sample_1_sample_data_1():

    maxBalls = 20
    maxRadius = 5.0
    td = ballcomp_sample_data_1()

    vh = ViolationHandler1(maxBalls,maxRadius)
    bc = BallComp(maxBalls,maxRadius,vh,True)
    
    for t in td:
        bc.conduct_decision(t)
    return
    

def test__BallComp__sample_1_sample_data_2():

    maxBalls = 4
    maxRadius = 0.2
    td = ballcomp_sample_data_2()
    vh = ViolationHandler1(maxBalls,maxRadius)
    bc = BallComp(maxBalls,maxRadius,vh,2)

    for t in td:
        bc.conduct_decision(t)

    print("********************")

    print("BALLS ", len(bc.balls))
    for k,v in bc.balls.items():
        print("k ",k)
        print(v)
        print()
    return

def test__BallComp__sample_1_sample_data_3():

    maxBalls = 5
    maxRadius = .5
    td = ballcomp_sample_data_3()
    vh = ViolationHandler1(maxBalls,maxRadius)

    # TODO: delete k
    bc = BallComp(maxBalls,maxRadius,vh,True)
    print("TD ",td.shape)

    for t in td:
        bc.conduct_decision(t)

    print("********************")
    print("BALLS ", len(bc.balls))
    for k,v in bc.balls.items():
        print("k ",k)
        print(v)
        print()
    return

def test__BallComp__sample_1_sample_data_4():

    maxBalls = 5
    maxRadius = 20.0

    vh = ViolationHandler1(15,80.0)

    filePath = "indep/ballcomp_sample_data_4.txt"
    ms = MessageStreamer(filePath,readMode = 'r')
    bc = BallComp(maxBalls,maxRadius,vh,2)

    q = 20
    s = 0
    while ms.stream() and q > 0:
        for t in ms.blockData:
            if bc.conduct_decision(t) != -1:
                s += 1
        q -= 1

    print("********************")
    print("BALLS ", len(bc.balls))
    s_ = 0
    for k,v in bc.balls.items():
        print(v)
        print()
        s_ += v.data.newData.shape[0]
    print("*********************")
    print("number of ball points {} actual {}".format(s_,s))
    return

if __name__ == "__main__":
    test__BallComp__sample_1_sample_data_4()



from .poly_struct import *
from .numerical_extras import *

# TODO: negative exponents
def int_split(ex,ratio):
    '''
    splits an integer into two values;
    the magnitude of the first value is determined
    by the float `ratio`
    '''

    # case: 0
    if ex == 0: return [0,0]
    p1 = int(round(ex * ratio,0))
    return (p1, ex - p1)

def modify_candidate_exponent(p1,modDirection):
    '''
    :param modDirection: [exponent,delta]
    '''
    v = deepcopy(p1.v)
    index = -1
    for (i,v_) in enumerate(v):
        if round(v_[1],5) == round(modDirection[0],5):
            index = i
            break
    assert index != -1, "invalid exponent for modification"
    v[index,1] += modDirection[1]
    return 

def remainder_to_target(p1,p2,P,targetPIndex):
    '''
    '''
    q = P.v[targetPIndex,1]
    l = valid_existing_candidate_elements_for_exponent(p1,p2,q)
    c = multiple_pairs_to_CEPoly(p1,p2,l)
    return P.v[targetPIndex,0] - c.v[0,0]

def valid_existing_candidate_elements_for_exponent(p1,p2,targetExp):
    '''
    gets all pairs of elements (e1,e2) such that e1 in `p1`,e2 in `p2`
    and pow(e1 * e2) = `targetExp`. 

    :return: list<(index of p1 multiple,index of p2 multiple)>
    '''

    a = []
    # get all possible
    for (i,v1) in enumerate(p1.v):
        for (j,v2) in enumerate(p2.v):
            if round(v1[1] + v2[1] - targetExp,5) == 0.:
                a.append((i,j))
    return a

def multiple_pairs_to_CEPoly(p1,p2,iv):
    '''
    '''

    v = []
    for x in iv:
        pe1,pe2 = p1.v[x[0]],p2.v[x[1]]
        sx = (pe1[0] * pe2[0],pe1[1] + pe2[1])
        v.append(sx)
    return CEPoly.from_dict(merge_duplicates_cevector(v))

def possible_new_elements_of_factor_for_exponent(f,p,feb,peb,P,t):
    '''
    determines the possible exponents of a new element of factor `f`
    to produce a polynomial P' with 
                `power_sequence(P') subset_in power_sequence(P')`.

    Possible exponents for `f` that when multiplied with `p` takes a lower
    priority than 

    :param f: polynomial factor to add a new element to
    :param p: non-mutable factor p to solve for `fp=P`
    :param feb: exponential bounds for factor f
    :param peb: exponential bounds for factor p
    :param P: polynomial to be factored by `f`
    :param t: int|float, target exponent
    '''

    c1 = []
    fp = set([vp[1] for vp in f.v])
    sp = set([vp[1] for vp in P.v])

    for x in range(int(feb[0]), int(feb[1]) + 1):
        # case: x already in f
        if x in fp:
            continue

        # case:
        x2 = t - x

        # case: needed x2 falls out of the exp. bounds of p1
        if x2 > peb[1] or x2 < peb[0]:
            continue

        # determine if multiplying exponent x2 with any in p2 produces an
        # exponent not in P
        s = set([x + p.v[i,1] for i in range(len(p.v))])

        # case: produces exponent not in P, possibility has low priority
        if len(s - sp) != 0:
            c1.append(x)
            continue
        else:
            c1.insert(0,x)
    return c1

def select_median_in_sequence(s,m=0.5):
    assert m >= 0. and m <= 1., "invalid median ratio"
    l = int((len(s) - 1) * m)
    return s.pop(l)

def median_sort(s,m= 0.5):
    s_ = []
    while len(s) > 0:
        s_.append(select_median_in_sequence(s,m))
    return s_

# TODO: rational numbers
# TODO: make sure bounds for multiple is accurate
def decisive_guess_sequence_experimental(c,complementFactor,remainingInsertions):
    if remainingInsertions == 0:
        if c % complementFactor:
            return [], False
        return [int(c / complementFactor)], True

    x = int(c / complementFactor)
    return [i for i in range(1,x + 1)], True

def decisive_guess_sequence(c,complementFactor):
    x = int(c / complementFactor)
    d = -1 if x >= 0 else 1
    return [i for i in range(x,0,d)] 

# TODO: negative exponents
class PolyFactorEst:
    '''
    '''

    def __init__(self,p,verbose=False):
        assert type(p) == CEPoly, "invalid polynomial"
        self.p = p
        self.plabels = set()
        self.p1,self.p2 = None,None
        self.p1eb,self.p2eb = None,None

        self.initial_bounds_multiples = []
        self.ibm_cache = []

        self.cache = []
        self.tmpCache = []
        self.guessCount = 0
        self.elimCount = 0 

        self.pindex = None
        self.f1,self.f2 = None,None

        self.stat = True
        self.verbose = verbose
        return

    def guess(self,median=0.5):
        while self.stat: 
            self.guess_iteration()
            self.set_factor_guess() 

    def guess_iteration(self):
        while self.guess_insertion():
            continue
        return

    def guess_insertion(self):

        if not self.stat: return False

        if type(self.f1) != type(None):
            if self.verbose: print("solved")
            return False

        if len(self.cache) == 0:
            if self.verbose: print("not solved")
            return False

        # pop element from cache
        (self.p1,self.p2) = self.cache.pop(0)
        (a1,a2) = self.analyze_factor_pair()
        if self.verbose: print("analysis: {} {}".format(a1,a2))

        # case: invalid factor-pair guess
        if not a2:
            self.elimCount += 1
            return True

        # case: no more remaining non-null powers for P;
        #       check for .
        if a1 == -1:
            if str(self.p - (self.p1 * self.p2)) == '':
                self.f1 = deepcopy(self.p1)
                self.f2 = deepcopy(self.p2)
            return True

        x = self.guess_insertion_candidates(1,a1)
        if x == 0:
            self.guess_insertion_candidates(0,a1)
        if self.verbose:
            print("UDPATE")
            for t in self.tmpCache:
                print(str(t[0]))
                print(str(t[1]))
                print("-@-@-@-@-@")
            print("============")
        self.guessCount += len(self.tmpCache)
        self.cache.extend(self.tmpCache)
        self.tmpCache = []
        return True

    def set_exponential_bounds(self,p1eb,p2eb):
        '''
        used in initial hypothesis
        '''
        assert type(p1eb) is tuple and len(p1eb) == 2 and p1eb[0] <= p1eb[1], "polynomial 1 exponential bounds"
        assert type(p2eb) is tuple and len(p2eb) == 2 and p2eb[0] <= p2eb[1], "polynomial 2 exponential bounds"
        self.p1eb = p1eb
        self.p2eb = p2eb

    def initial_bounds_hypothesis(self,median=0.5):
        # case: P has length 1
        if len(self.p.v) < 2:
            #self.p.v[0]
            return

        # case: P has length >= 2
            # choose the multiples
        m = median_sort(all_multiple_pairs(self.p.v[0,0]),median)
        m2 = median_sort(all_multiple_pairs(self.p.v[-1,0]),median)

        self.ibm_cache = [None,[]]
        self.initial_bounds_multiples = [m,deepcopy(m2)]
        self.set_factor_guess()

        if self.p1 * self.p2 == self.p:
            self.f1,self.f2 = self.p1,self.p2
            self.stat = False

    def set_factor_guess(self):
        '''
        declares each 
        '''
        if len(self.initial_bounds_multiples[0]) == 0:
            self.stat = False
            return False

        # case: second multiple-pair candidate list exhausted 
        if len(self.ibm_cache[1]) == 0:
            self.ibm_cache[0] = self.initial_bounds_multiples[0].pop(0)
            self.ibm_cache[1] = deepcopy(self.initial_bounds_multiples[1])

        mdn2 = self.ibm_cache[1].pop(0)
        x1 = np.array([(deepcopy(self.ibm_cache[0][0]),self.p1eb[1]),(mdn2[0],self.p1eb[0])])
        x2 = np.array([(deepcopy(self.ibm_cache[0][1]),self.p2eb[1]),(mdn2[1],self.p2eb[0])])
        self.p1 = CEPoly(x1)
        self.p2 = CEPoly(x2)
        self.cache.append((deepcopy(self.p1),deepcopy(self.p2)))
        return True

    def guess_insertion_candidates(self,pid,pindex):
        ex = self.p.v[pindex,1]
        (exp1,coeff) = self.new_max_exponent_w_complement_for_factor(pid,ex)
        
        if coeff == None:
            return 0

        # guess the possible coefficients
        d = decisive_guess_sequence(deepcopy(self.p.v[pindex,0]),coeff)
        
        # add candidates to tmp cache
        q = [deepcopy(self.p1),deepcopy(self.p2)]
        for x in d:
            # make the CEPoly
            
            px = CEPoly.from_dict({exp1: x})
            
            # add the CEPoly
            qx = deepcopy(q)
            qx[(pid + 1) % 2] = qx[(pid + 1) % 2] + px
            self.tmpCache.append(qx)
        return len(d)
    
    def new_max_exponent_w_complement_for_factor(self,pid,te):
        '''
        :param pid:
        :param te:
        :return: (exponent for factor `pid`, coefficient of complement)
        '''
        new_elements = possible_new_elements_of_factor_for_exponent(self.p1,self.p2,self.p1eb,self.p2eb,self.p,te)\
                if pid else possible_new_elements_of_factor_for_exponent(self.p2,self.p1,self.p2eb,self.p1eb,self.p,te)

        # iterate through and get the element with the 
        for n in new_elements:
            n2 = te - n
            c = self.p2.c_of_p(n2) if pid else self.p1.c_of_p(n2)
            if c != None:
                return (n,c)
        return (None,None)

    def analyze_factor_pair(self):
        '''
        perform analysis of p1 and p2. 

                - First output:
        Determine the first element from the max in `P` that requires
        insertions.

                - Second output: 
        If `p1 * p2` produces a polynomial with exponents that parent 
        polynomial `P` does not own, and there does not exist possible
        new insertions into either `p1` and `p2` that can cancel out 
        the excess exponents, output `False`. Otherwise, `True`. 
        '''
        ###
        #print("analyzing")
        #print("P1: {}",str(self.p1))
        #print("P2: {}",str(self.p2))
        #print("----------------------")
        ###

        if not self.excess_element_analysis():
            return None,False

        if not self.unequal_element_analysis():
            return None,False

        px = self.p1 * self.p2
        for (i,q) in enumerate(self.p.v):
            c2 = px.c_of_p(q[1])
            if c2 == None: c2 = 0.

            if round(q[0] - c2,5) != 0.:
                return (i,True)
        return -1,True

    def excess_element_analysis(self):
        es = (self.p1 * self.p2).exponential_subtract(self.p)
        for s in es:
            # possible elements 
            q1 = len(possible_new_elements_of_factor_for_exponent(self.p1,self.p2,self.p1eb,self.p2eb,self.p,s))
            q2 = len(possible_new_elements_of_factor_for_exponent(self.p2,self.p1,self.p2eb,self.p1eb,self.p,s))
            if q1 + q2 == 0: 
                return False
        return True

    def unequal_element_analysis(self):
        p3 = self.p1 * self.p2
        exp = []
        for v_ in p3.v:
            c = self.p.c_of_p(v_[1])
            if c - v_[0] != 0.:
                exp.append(v_[1])

        for s in exp:
            q1 = len(possible_new_elements_of_factor_for_exponent(self.p1,self.p2,self.p1eb,self.p2eb,self.p,s))
            q2 = len(possible_new_elements_of_factor_for_exponent(self.p2,self.p1,self.p2eb,self.p1eb,self.p,s))
            if q1 + q2 == 0:
                return False
        return True

    def check_valid_guess(self):
        return -1

    def poly_type(self):
        self.plabels = set()
        for x in self.p.v:
            # no more searching needed.
            if self.plabels == {complex,"q"}: break

            if type(x) == complex:
                self.plabels |= {complex}
            if int(x) != x:
                self.plabels |= {"q"}
        return


from morebs2 import message_streamer
import unittest

'''
python -m morebs2.tests.message_streamer_test  
'''
class TestMessageStreamerClass(unittest.TestCase):

    def test_MessageStreamer__open_file(self):
        fp = "morebs2/indep/text1.txt"
        fp2 = "morebs2/indep/tree_bark_1.jpg"
        fp3 = "morebs2/indep/text2.txt"

        try:
            q = message_streamer.MessageStreamer(fp)
            q2 = message_streamer.MessageStreamer(fp2)
            q3 = message_streamer.MessageStreamer(fp3)
            assert q.mav == "tex" and q2.mav != q.mav, "incorrect file recognition"
            print("ENCODING OF ", q3.textEncoding)

            q.end_stream()
            q2.end_stream()
            q3.end_stream()
        except:
            raise ValueError("error opening files")

    """
    """
    def test_MessageStreamer__stream_block__text(self):
        fp = "morebs2/indep/text1.txt"
        q = message_streamer.MessageStreamer(fp)
        q.stream_block()
        assert len(q.blockData) > 0 and len(q.blockData) <= message_streamer.DEFAULT_STREAM_BLOCK_SIZE, "invalid block data size"

        print("\n\t**DISPLAY::stream_block__text")
        for l in q.blockData:
            print(l)
        q.end_stream()

    def test_MessageStreamer__stream_block__jpg(self):
        fp = "morebs2/indep/tree_bark_2.jpg"
        q = message_streamer.MessageStreamer(fp)
        q.stream_block()
        assert len(q.blockData) == message_streamer.DEFAULT_STREAM_BLOCK_SIZE, "invalid block data size, got {}".format(len(q.blockData))

        print("\n\t**DISPLAY::stream_block__jpg")
        for l in q.blockData:
            print(l)
        q.end_stream()

    def test_MessageStreamer__stream_block__csv(self):
        fp = "morebs2/indep/s.txt"
        q = message_streamer.MessageStreamer(fp,readMode="r")
        q.stream_block()

        print("\n\t**DISPLAY::stream_block__csv")
        for l in q.blockData:
            print(l)
        q.end_stream()

if __name__ == '__main__':
    unittest.main()



import os
import glob
import csv
from collections import defaultdict
from operator import add,sub,mul,truediv,floordiv
import numpy as np 

WHITE = (255,255,255)
BLACK = (0,0,0)
BLUE = (0,0,255)
REDGREENOREGON = (123,321,231)

DEFAULT_TRAVELLING_HOP = 0.05
DEFAULT_SCREEN_DIM = (1300,750)
DEFAULT_POINT_CACHE_SIZE = 10


NUMERICAL_TYPES = {int,float,\
    np.int8,np.int16,np.int32,np.int64,\
    np.float16,np.float32,np.float64,\
    complex,np.complex64,np.complex128}

# TODO:
def clear_and_make_directory(dirPath):
    """
    clear_and_make_directory

    :param dirPath: directory path
    :type dirPath: str or None
    :return: None
    :rtype: None
    """

    ##dirr = os.path.dirname(fullPath)
    if not os.path.isdir(dirPath):
        # make directory
        os.mkdir(dirPath)
    else:
        # TODO: clear directory contents
        files = glob.glob(dirPath + "/*")
        for f in files:
            try:
                os.remove(f)
            except:
                print("probably is folderonos")

def make_csv_file(filePath, columnLabels):
    with open(filePath, 'w', newline='') as csvfile:
        writer = csv.writer(csvfile, delimiter=',', quotechar='|', quoting=csv.QUOTE_MINIMAL)
        writer.writerow(columnLabels)

def is_valid_color(c):
    assert type(c) is tuple, "invalid color type"
    for c_ in c: assert type(c_) in [float,int], "invalid type for colores de la vegasvalueorios"

##################### methods for dictionaries 

def std_invert_map(m):
    assert type(m) in {dict,defaultdict}

    q = {}
    for k,v in m.items():
        if v in q:
            q[v].append(k)
        else:
            q[v] = [k]
    return q

def invert_map__seqvalue(m): 
    assert type(m) in {dict,defaultdict}

    q = defaultdict(list) 
    for k,v in m.items(): 
        for v_ in v: 
            q[v_].append(k) 
    return q 

def numberdict_op(d1,d2,f=sub):
    K = set(d1.keys()) | set(d2.keys()) 
    
    d3 = defaultdict(int) 
    for k in K:
        x1 = d1[k] if k in d1 else 0 
        x2 = d2[k] if k in d2 else 0 
        d3[k] = f(x1,x2) 
    return d3 

def equal_intdicts(d1,d2): 
    K = set(d1.keys()) | set(d2.keys()) 

    for k in K: 
        if d1[k] != d2[k]: return False 
    return True 



from .line import *
from collections import defaultdict
from copy import deepcopy

ndim_point_distance = lambda p1, p2: np.sum([np.sqrt((p1 - p2)**2)])


def tie_sets_for_vector(vector):
    """Iterate through sorted vector for sets of
    indices. Tie-sets are represented as a sequence of
    (startIndex,endIndex) pairs due to vector
    being ordered.
    """
    if len(vector) == 0: return []

    l = vector.shape[0]

    tss = []
    ts = [0]
    tsv = vector[0]
    i = 1
    while i < l:

        if ts == []:
            if i + 1 == l:
                i += 1
                continue

            if vector[i] == vector[i + 1]:
                ts = [i, i + 1]
                tsv = vector[i]
                i += 2
            else:
                i += 1

        else:
            # pass
            if vector[i] != tsv:

                if ts == [0]:
                    ts = []
                    continue

                tss.append(ts)
                ts = []
            else:
                ##print("NN")
                ts = [ts[0], i]
            i += 1

    if ts != []: tss.append(ts)

    return tss

def sorted_vector_value_range_search(vector, value):

    head, stat = sorted_vector_value_search(vector, value)

    if not stat:
        return head,stat
    q = head + 1
    if q >= vector.shape[0]: return [head,head], stat

    while vector[q] == value:
        q += 1
        if q >= vector.shape[0]: break
    return [head,q - 1], stat

def sorted_vector_value_search(vector, value, irange = None):
    """
    Iterates through a sorted vector for `value`.
    Outputs the "head" index of the value if it exists
    """
    s = np.where(vector == value)[0]

    if len(s) == 0:
        return None,False
    return s[0],True

class PointSorter:
    """
    class can be used by one of these ways: 
    
        (1) instantiate class w/ vector of data, and sort data using
    method<sort_it>, 
        (2) instantiate class w/ empty vector, call method<insert_point>
    to add to data. The sorted vector point data will be variable<newData>.
    
    """

    def __init__(self,data):
        assert is_2dmatrix(data), "invalid data"

        # initial data variable (static)
        self.data = data
        self.newData = np.copy(data)
        self.pointMean,self.pointVar = None,None

        # key: -1 for min, 1 for max
        # value: [0] is vector, [1] is accomodating float
        '''
        variable contains referential extremum pertaining to a reference point.
        Extreme points are calculated based on their magnitudes to reference point.
        '''
        self.extremumData = defaultdict(tuple)
        self.deltaCache = defaultdict(list)
        self.sort_it()
        return

    def clear_delta_cache(self):
        self.deltaCache = defaultdict(list)
        self.deltaCache[-1] = np.empty((0,self.pointMean.shape[0]))
        self.deltaCache[1] = np.empty((0,self.pointMean.shape[0]))

    def sort_it(self):
        if self.newData.shape[0] <= 1:
            return
        self.sort_at_column(0, [0, self.newData.shape[0] - 1])

    def sort_at_column(self, columnIndex, submatrixIndices):

        if columnIndex >= self.newData.shape[1]:
            print("doncesorded")
            return

        # sort submatrix at column
        submatrix = self.newData[submatrixIndices[0]:submatrixIndices[1] + 1]
        q = submatrix[:,columnIndex]

        indices = np.argsort(q)
        newSM = submatrix[indices]
        self.newData[submatrixIndices[0]:submatrixIndices[1] + 1] = newSM

        # check for tie-sets in submatrix
        newV = newSM[:,columnIndex]
        ts = tie_sets_for_vector(newV)
        if len(ts) == 0:
            return
        else:
            for t in ts:
                # calibrate indices
                ci = [submatrixIndices[0] + t[0], submatrixIndices[0] + t[1]]
                # sort
                self.sort_at_column(columnIndex + 1, ci)
        return

    ###---------------------start: point existence --------------------------------------------------------

    def vector_exists(self, vector):
        """
        :return: index of vector that it exists at, -1 if does not exist.
        :rtype: int
        """
        return self.does_vector_exist_(vector, 0, [0, self.newData.shape[0]])

    def does_vector_exist_(self, vector, columnIndex, sampleIndices):

        # case: possible ties due to no more columns, return the head
        if columnIndex >= vector.shape[0]:
            return sampleIndices[0]

        c = self.newData[sampleIndices[0]: sampleIndices[1] + 1][:,columnIndex]
        irange,stat = sorted_vector_value_range_search(c, vector[columnIndex])

        # case: column value does not exist, !vector
        if not stat: return -1

        # case: single sample
        if irange[1] - irange[0] == 0:
            return irange[0] + sampleIndices[0]
        else:
        # multiple samples, iterate on next column
            # new sample indices are irange + calibration ()
            nsi = [sampleIndices[0] + irange[0], sampleIndices[0] + irange[1]]
            return self.does_vector_exist_(vector, columnIndex + 1, nsi)

    ###---------------------end: point existence --------------------------------------------------------

    ###---------------------start: point insertion/deletion --------------------------------------------------------

    def insert_point(self,p):

        # case: empty data, insert
        if self.newData.shape[0] == 0:
            self.newData = np.vstack((self.newData, p))
            return

        # case: non-empty data
        index = 0
        insertIndex = -1
        remVar = np.arange(self.newData.shape[0])
        while index < self.newData.shape[1]:
            v = p[index]
            insertIndex, remVar = self.data_index_for_value_at_column(v,index,remVar)
            if remVar == []:
                break
            index += 1

        # insert at index
        self.insert_at(p, insertIndex)

    '''
    return:
    - index::int, (tie set vector)::bool
    '''
    def data_index_for_value_at_column(self, v, c, possibleIndices):
        # case: zero choices, output -1
        if len(possibleIndices) == 0: return -1

        x = None
        for (i,j) in enumerate(possibleIndices):
            if v <= self.newData[j,c]:
                x = i
                break

        # case: found index, check for ties
        if x != None:
            # subcase: no ties
            if v != self.newData[possibleIndices[x],c]:
                return possibleIndices[x],[]

            # subcase: ties, collect all ties
            l = len(possibleIndices)
            q = [possibleIndices[x]]
            q2 = self.newData[possibleIndices[x],c]
            for j in range(x + 1, l):
                if self.newData[possibleIndices[j],c] == q2:
                    q.append(possibleIndices[j])
                    continue
                break

            return possibleIndices[x],q

        # case: after possible indices, no ties
        return possibleIndices[-1] + 1, []

    def insert_at(self, p, i):
        self.newData = np.concatenate((self.newData[:i], [p], self.newData[i:]))

    def delete_point(self,p):
        i = self.vector_exists(p)

        # case: point does not exist
        if i == -1: return
        self.newData = np.concatenate((self.newData[:i], self.newData[i + 1:]))
        return

    ###---------------------end: point insertion/deletion --------------------------------------------------------

    ###---------------------start: point update by timestamp -----------------------------------------------------

    def log_delta_cache(self,p,additionBool):
        assert additionBool in [1,-1], "invalid addition bool"
        q = self.deltaCache[additionBool]

        if len(q) == 0:
            q = np.array([p])
        else:
            q = np.vstack((q,p))
        self.deltaCache[additionBool] = q

    def update_points_from_cache(self, referencePoint = None):
        '''
        Updates points from delta cache and clears it.
        '''

        # update mean and var
            # update +1
        p = 1
        q = self.delta_points(p)
        self.update_points(q,p)

            # update -1
        p = -1
        q = self.delta_points(p)
        self.update_points(q,p)

        # update extremum
        if type(referencePoint) != type(None):
            self.update_extremum(referencePoint)
        #
        self.deltaCache[-1] = []
        self.deltaCache[1] = []

    def update_points(self,additionalPoints,additionBool):

        if additionBool == -1:
            self.safe_deletion_check(additionalPoints)

        # update measures
        self.update_mean(additionalPoints,additionBool)
        self.update_var(additionalPoints,additionBool)

        # update point data
        f = self.insert_point if additionBool == 1 else self.delete_point
        for i in range(additionalPoints.shape[0]):
            f(additionalPoints[i])

    def delta_points(self,additionBool):
        q = self.deltaCache[additionBool]

        # case: empty
        if type(q) is list:
            return np.empty((0,self.newData.shape[1]))
        else:
            return q

    ###---------------------end: point update by timestamp -----------------------------------------------------

    ###---------------------start: measure update functions

        ####----- methods below perform calculations on entire data `newData`

    def calculate_mean(self):
        self.pointMean = np.mean(self.newData,0)

    def calculate_variance(self):
        self.pointVar = np.var(self.newData,0)

    def calculate_extremum(self,referencePoint):
        d = [ndim_point_distance(referencePoint,p) for p in self.newData]
        imin,imax = np.argmin(d),np.argmax(d)

        self.extremumData[-1] = (self.newData[imin], d[imin])
        self.extremumData[1] = (self.newData[imax], d[imax])

    @staticmethod
    def update_meanbasedvalue(meanBasedValue, pointShape,additionalPoints,additionBool = 1):
        assert is_vector(meanBasedValue) or type(meanBasedValue) is type(None), "invalid mean based value"
        assert is_valid_point(pointShape),"invalid point shape"
        assert is_2dmatrix(additionalPoints), "invalid additional points"
        assert additionalPoints.shape[1] == pointShape[1], "invalid shape for additional points"

        # case: no additional points
        if additionalPoints.shape[0] == 0: return meanBasedValue

        if type(meanBasedValue) is type(None):
            meanBasedValue = np.zeros((pointShape[1]))
        assert meanBasedValue.shape[0] == pointShape[1], "invalid shape for meanBasedValue w/ {}".format(pointShape)

        # case: points are to be subtracted
        if additionBool == -1:
            assert additionalPoints.shape[0] <= pointShape[0], ""

        # prev sum
        s = float(pointShape[0]) * meanBasedValue
        # additional sum
        s2 = np.sum(additionalPoints,axis=0)

        if additionBool == 1:
            # new average
            l = additionalPoints.shape[0] + pointShape[0]
            s3 = (s + s2) / float(l)
        else:
            # new average
            l = pointShape[0] - additionalPoints.shape[0]
            s3 = (s - s2) / float(l)

        return np.round(s3,5)

    def update_mean(self, additionalPoints,additionBool):
        mbs = PointSorter.update_meanbasedvalue(self.pointMean, tuple(self.newData.shape),additionalPoints,additionBool)
        self.pointMean = mbs

    def update_var(self,additionalPoints,additionBool):
        mbs = PointSorter.update_meanbasedvalue(self.pointVar, tuple(self.newData.shape),additionalPoints,additionBool)
        self.pointVar = mbs

    def safe_deletion_check(self,points):
        for p in points: assert self.vector_exists(p)

    def update_extremum(self,referencePoint):
        '''
        Updates extremum based on euclidean distance of reference point to
        points in cache.

        Call this method after `deltaCache` has been updated
        '''
        # addition
        self.update_extremum_(referencePoint,self.deltaCache[1])

        # subtraction
        if type(self.deltaCache[-1]) != list:

            self.calculate_extremum(referencePoint)

    def update_extremum_(self,referencePoint, additionalPoints):
        # case: no additional points
        if len(additionalPoints) == 0: return
        if additionalPoints.shape[0] == 0: return

        q = 0

        # case: empty
        if self.extremumData[-1] == () and self.extremumData[1] == ():

            p = additionalPoints[0]
            d = ndim_point_distance(referencePoint,p)
            self.extremumData[-1] = (p,d)
            self.extremumData[1] = (p,d)
            q = 1

        # update remainder
        for i in range(q,additionalPoints.shape[0]):
            p = additionalPoints[i]
            d = ndim_point_distance(referencePoint,p)

            if d < self.extremumData[-1][1]:
                self.extremumData[-1] = (p,d)

            elif d > self.extremumData[1][1]:
                self.extremumData[1] = (p,d)

        return

    def extreme(self,maxBool):
        assert maxBool in [-1,1]
        return deepcopy(self.extremumData[maxBool])

    ###---------------------end: measure update functions



from .ball_volume_estimators import *
from .ball_operator import *
from .violation_handler import *

def dec1(u,t,p):
    x1 = 1 - u / t
    x2 = t / p
    return (x1 + x2) / 2.0

def dec2(b1,b2):
    return b1/b2 * 1/2


# do violation recommendation
# verbose
# arg. history (max balls, max radius)
class BallComp:

    MIN_LENGTH_RATIO = 0.01

    def __init__(self, maxBalls,maxRadius,vh1,verbose = 0):
        assert maxRadius > 0.0 and maxBalls > 0, "invalid args. for BallComp"
        self.maxBalls = maxBalls
        self.maxRadius = maxRadius
        self.balls = {} # int:idn -> Ball
        self.pointMem = None # (label of last point added, last point added)
        self.ballNeighborsUpdate = None

        assert verbose in [2,1,0]
        self.verbose = verbose
        self.dve = DisjunctionVolumeEstimator()
        self.ballCounter = 0

        self.vh = vh1
        self.terminateDelta = False

        self.ts = 0 # timestamp

        # keeps track of argument history
        # each element is (value,timestamp)
        self.bah = []
        self.rah = []

    #### start: main function --------------------------------------------------

    """
    """
    def conduct_decision(self,p):

        if self.terminateDelta:
            print("NO MORE ADD")
            return -1

        self.dve.clear_cache()
        if self.verbose == 2:
            print("\t* pre-decision measures")
            self.summarize_volume_measures()
        d1 = self.decision_1_score(p)
        d2 = self.decision_2_score()
        self.dve.clear_cache()

        # case: make new ball
        if d2 < d1[2]:
            if self.verbose >= 1:
                print("-- choose decision 2")
            b = Ball(p,self.ballCounter)
            b.add_element(p)
            b.radius = BallComp.MIN_LENGTH_RATIO * self.maxRadius
            self.ballCounter += 1

            ## TODO: add ball here
            self.add_ball(b)
            self.load_recommendation((2,p,b.idn))
            if self.verbose >= 1:
                print("-------------------------------------------")
            return

        # case: add to present ball
        self.balls[d1[1]].add_element(p)
        if self.verbose >= 1:
            print("--- decision 1 ball")
            print(self.balls[d1[1]])
            print("----")

            # update neighbors
        self.update_neighbors_of_ball(d1[1])

            # subcase: no merge
            ## update ball volume and its neighbor 2-intersection volumes
        if d1[0] == 1:
            if self.verbose >= 1: print("-- choose decision 1-no merge")
            self.update_target_ball_volume(d1[1])
            self.update_target_ball_2int(d1[1])
            nbi = d1[1]
        else:
                # subcase: merge
                ## merge balls into one
            print("-- choose decision 1-merge")
                ##bs = self.dataless_ball_copies(self.balls[d1[1]].neighbors | {d1[1]})
            x = self.balls[d1[1]].neighbors | {d1[1]}
            bs = [self.balls[x_] for x_ in x]
            ball0 = BallComp.merge_ball_list(bs)
            ball0.idn = int(self.ballCounter)
            self.ballCounter += 1
                ## delete all values in self.balls[d1[1]].neighbros | {d1[1]}
            self.remove_ballset(self.balls[d1[1]].neighbors | {d1[1]})
                ## add ball0
            self.add_ball(ball0)
            nbi = ball0.idn

        self.load_recommendation((1,self.balls[nbi].radius,nbi))
        print("-------------------------------------------")

    def summarize_volume_measures(self):
        print(self.dve.ballVolumes)
        print("---")
        print(self.dve.d)
        print("\t********")

    #### end: main function --------------------------------------------------

    #### start: method requirements for decision 1
    ########### start: point add functions

    def update_neighbors_of_ball(self,idn):
        '''
        calculates the new neighbor set N1 of the bl'th ball that used to have the neighbor
        set N0. Then update the `neighbors` variable for all affected neighbors of the
        bl'th ball.

        :param idn:
        :type idn: int
        '''
        q = self.balls[idn].neighbors
        self.balls[idn].neighbors = self.neighbors_of_ball(idn)

        self.ballNeighborsUpdate = (idn,q,deepcopy(self.balls[idn].neighbors))
        self.update_ball_neighbors_var(self.ballNeighborsUpdate[0],\
            self.ballNeighborsUpdate[1],self.ballNeighborsUpdate[2])

    def update_ball_neighbors_var(self,idn,n0,n1):
        '''
        updates the neighbors of each ball after a ball has a radius change by the rule: 
        - positive difference set N1 - N0: adds `idn` to these balls' neighbors.
        - negative difference set N0 - N1: subtracts `idn` from .these balls' neighbors.
        '''

        pd = n1 - n0
        nd = n0 - n1
        for p in pd:
            self.balls[p].neighbors = self.balls[p].neighbors | {idn}
        for n in nd:
            self.balls[n].neighbors = self.balls[n].neighbors - {idn}
        return

    def revert_update_neighbors_of_ball(self):
        self.update_ball_neighbors_var(self.ballNeighborsUpdate[0],\
            self.ballNeighborsUpdate[2],self.ballNeighborsUpdate[1])
        self.ballNeighborsUpdate = None
        return

    #@
    def neighbors_of_ball(self,idn):
        b = self.balls[idn]
        n = set()
        for k,v in self.balls.items():
            if idn == k: continue
            if v.is_neighbor(b): n.add(k)
        return n

    def add_point_to_ball(self,p,idn):
        self.balls[idn].add_element(p)
        return

    def ball_label_for_point(self, p):
        '''
    determines the ball idn for point based on minumum
    euclidean point distance
        '''
        bc = np.array([(euclidean_point_distance(b.center,p),b.idn)\
            for b in self.balls.values()])

        if len(bc) == 0:
            return -1
        i = np.argmin(bc[:,0])
        return int(bc[i,1])

            ###### special cases for adding point to ball

    def ball_label_for_point__qualify_radius(self,p):
        """
        Determines ball label for point based on balls' current radii, as they say,
        "variantate` de las labelovos de los qualifacados"

        :param p: vector
        :type p: np.array
        :return: los labelovas
        :rtype: int
        """
        bc = np.array([(euclidean_point_distance(b.center,p),b.idn)\
            for b in self.balls.values()])

        if len(bc) == 0:
            return -1

        # sort by ascending euclidean point distance
        order = np.argsort(bc[:,0])
        for o in order:
            if bc[o,0] <= self.maxRadius:
                return self.balls[bc[o,1]].idn
        return -1

    def restrict_add_point(self,p):
        return -1

    ########### end: point add functions

    def dataless_ball_copies(self,indices):
        '''
        returnia the dataless ball copies of that indices

        :type indices: iter
        :return: returnia
        :rtype: list<listinia de ballco>
        '''
        return [Ball.dataless_copy(self.balls[i]) for i in indices]

    '''
    assumes all balls are neighbors
    '''
    @staticmethod
    def merge_ball_list(bs):
        if len(bs) == 0: return None

        # sort ball set in ascending distance from .ball-set mean
        q = np.array([bs_.center for bs_ in bs])
        m = np.mean(q,axis = 0)
        d = [euclidean_point_distance(bs_.center,m) for bs_ in bs]
        indices = list(np.argsort(d))

        # merge in that order
        i = indices.pop(0)
        b_ = bs[i]

        while len(indices) > 0:
            i = indices.pop(0)
            b_ = b_ + bs[i]
        return b_

    #### end: method requirements for decision 1

    #### start: decision function 1

    def pre_decision_1_(self,p,idn):

        # add point to ball
        self.add_point_to_ball(p,idn)

        # update its neighbors
        if self.verbose >= 1:
            print("\t\tprevious neighbors:\n\t",vector_to_string(sorted(self.balls[idn].neighbors)))

        self.update_neighbors_of_ball(idn)

        if self.verbose >= 1:
            print("\t\tnew neighbors:\n\t",vector_to_string(sorted(self.balls[idn].neighbors)))

        # update target ball volume
        self.update_target_ball_volume(idn)

        # update target ball 2-int
        self.update_target_ball_2int(idn)

        return

    def update_target_ball_volume(self,idn):
        ###print("logging ball volume for {}: {}".format(idn,self.balls[idn].radius))
        self.dve.log_ball_volume(self.balls[idn])

    def update_target_ball_2int(self,idn):
        q = self.balls[idn].neighbors
        for q_ in q:
            b2 = self.balls[q_]
            ###print("LOGGING 2INT")
            self.dve.log_ball_volume_2intersection(self.balls[idn],b2)
        return

    def post_decision_1_(self,idn):
        # revert all changes made
            # target ball add point
        self.balls[idn].revert_add_point()
            # target ball neighbors
        self.revert_update_neighbors_of_ball()
        self.dve.revert_cache_delta(1)
        self.dve.revert_cache_delta(2)

    def decision_1_score(self,p):
        bl = self.ball_label_for_point(p)

        if self.verbose >= 1:
            print("\tSIM:\n\tdecision 1")
            print("\t\tadding point to: ",bl)
        # case: no balls to choose
        if bl == -1:
            return (1,bl,2.0)

        self.pre_decision_1_(p,bl)

        bs1 = self.balls[bl].neighbors | {bl}
        # simulate 1: no merge
        if self.verbose:
            print("estimating disjunction")
        x = True if self.verbose == 2 else False
        vu = self.dve.estimate_disjunction_at_target_ball(self.balls[bl].idn,x,500)
        vt = np.sum([ball_area(self.balls[i].radius,p.shape[0]) for i in bs1])
        vp = ball_area(self.maxRadius,p.shape[0]) * len(bs1)
        d1 = dec1(vu,vt,vp)
        if self.verbose >= 1:
            print("\t\t no-merge volume measures: ",vu,vt,vp)
            print("\t\t no-merge score: ", d1)

        # simulate 2: merge target ball w/ its neighbors
        ballSet = self.dataless_ball_copies(bs1)

        ball0 = BallComp.merge_ball_list(ballSet)
        vu = ball_area(ball0.radius,p.shape[0])
        vt = vu
        vp = ball_area(self.maxRadius,p.shape[0])
        d2 = dec1(vu,vt,vp)
        if self.verbose >= 1:
            print("\t\t merge volume measures: ",vu,vt,vp)
            print("\t\t merge score: ", d2)

        # choose the better option
        option = (1,bl,d1) if d1 <= d2 else (2,bl,d2)

        if self.verbose >= 1:
            print("\t\t decision 1 option: ",option)

        # revert changes
        self.post_decision_1_(bl)

        return option

    #### end: decision function 1

    #### start: decision function 2

    def decision_2_score(self):
        score = dec2(len(self.balls) + 1, self.maxBalls)

        if self.verbose >= 1:
            print("\tdecision 2 option:\n\t\tballs {} max balls {} score {}".format(len(self.balls),self.maxBalls,score))
        return score

    #### end: decision function 2

    #### start: decision 1 merge- ballset removal from .neighbors

    def remove_ballset(self,idns):

        s = set()
        # get affected neighbors from .ballset removal
        for idn in idns:
            s = s | self.balls[idn].neighbors

            # delete ball
            del self.balls[idn]

        # filter out all balls found in idns
        s = s - idns

        # remove labels of ballset from .affected neighbors
        self.delete_balls_from_affected_neighbors(idns,s)

        # remove all volume and intersection values in `dve` that contain
        # any label
        self.dve.delete_keyset(idns)

        return

    def delete_balls_from_affected_neighbors(self,bs,neighbors):
        for n in neighbors:
            self.balls[n].neighbors = self.balls[n].neighbors - bs
        return

    #### start: adjustment

    def add_ball(self,b):
        if self.verbose >= 1:
            print("\t * adding ball ", b.idn)

        self.balls[b.idn] = b
        self.update_ball_info(b.idn)
        return

    def update_ball_info(self,idn):
        self.update_neighbors_of_ball(idn)
        self.update_target_ball_volume(idn)
        self.update_target_ball_2int(idn)
        self.dve.clear_cache()

    """
    decision := (1,ball radius::(merged|non-merged),ball idn) | (2,p,ballIdn)
    """
    def load_recommendation(self,decision):
        assert decision[0] in {1,2}, "invalid decision"

        if decision[0] == 1:
            if self.verbose: print("$$ violation 1 check")
            res1 = self.vh.check_violation_1(decision[1],self.maxRadius)

            # perform ball-split
            if res1[0]:
                if self.verbose: print("-- max radius exceeded: splitting ball")
                self.fix_violation_1_by_ball_split(decision[2])

            ### TODO: add ball-split here as well
            elif not res1[0] and type(res1[1]) != type(None):
                # log old argument
                self.log_arg('r')

                # update `maxRadius`
                q = self.maxRadius + res1[1]
                if self.verbose: print("-- updating max radius: {}->{}".format(self.maxRadius,q))
                self.change_arg(('r',q))
                return
        else:
            if self.verbose: print("$$ violation 2 check")
            res2 = self.vh.check_violation_2(len(self.balls),self.maxBalls)
            self.fix_violation_2((decision[2],decision[1]),res2)
            return


    """
    splits balls by a greedy splitting scheme
    """
    def split_ball(self,idn):
        assert self.balls[idn].radius > self.maxRadius, "cannot split ball under radius"

        b = BallOperator(self.balls[idn])
        b.run_subball_split((self.maxRadius * 0.9,"literal"),"minimal",self.verbose)
        return b

    def save_split_ball(self,idn,bo):
        # save balls
        for s in bo.subballs:
            s.idn = self.ballCounter
            self.ballCounter += 1
            self.balls[s.idn] = s

        # update their info
        for s in bo.subballs:
            self.update_ball_info(s.idn)

        # delete ball
        self.remove_ballset({idn})
        return

    """
    violation 1 occurs when target ball needs to be split
    """
    def fix_violation_1_by_ball_split(self,bIdn):
        if self.verbose: print("*** Fixing violation 1 by split")
        bo = self.split_ball(bIdn)
        if self.verbose: print("Number of additional balls {}".format(len(bo.subballs) - 1))
        # check for violation of number of balls
        newNumberOfBalls = len(self.balls) - 1 + len(bo.subballs)
        if newNumberOfBalls > self.maxBalls:
            if self.verbose: print(" violation: {}".format(True))
            res2 = self.vh.check_violation_2(newNumberOfBalls,self.maxBalls)

            # terminate: cannot fix violation 1
            if res2[0]:
                if self.verbose: print("violation 1 cannot be fixed. terminate delta.")
                self.terminateDelta = True

                bo.revert_split()
            # do not terminate:
            else:
                # log old argument
                self.log_arg('b')

                # update `maxBalls`
                q = self.maxBalls + res2[1]
                self.change_arg(('b',q))

                if self.verbose: print("violation 1 -> new max balls: {}".format(q))

                # save split
                self.save_split_ball(bIdn,bo)

        # no violation
        else:
            self.save_split_ball(bIdn,bo)
        return

    """
    number of balls violation

    newBallInfo := (ball idn,new point)
    violation := (bool::violation,value)
    """
    def fix_violation_2(self,newBallInfo,violation):
        # case: max balls reached

        if violation[0]:
            if self.verbose: print("*** violation 2 cannot be fixed")
            # delete new ball
            ## new ball will not have any neighbors
            self.remove_ballset({newBallInfo[0]})
                ##del self.balls[newBallInfo[0]]
                ##self.dve.delete_keyset({newBallInfo[0]})

            # find a ball label for point
            l = self.ball_label_for_point__qualify_radius(newBallInfo[1])
            # ball found, add point to ball
            if l != -1:
                if self.verbose: print("found alternative ball @ {}".format(l))
                self.pre_decision_1_(newBallInfo[1],l)
            return
        # case: update maxBalls
        elif not violation[0] and type(violation[1]) != type(None):
            if self.verbose: print("*** violation 2 fixed")
            self.log_arg('b')
            q = self.maxBalls + violation[1]
            self.change_arg(('b',q))
            if self.verbose: print("new max number of balls: ",q)
            return

    def log_arg(self,arg):
        assert arg in {'b','r'}, "invalid argument"
        x = self.maxBalls if arg == 'b' else self.maxRadius

        if arg == 'b':
            self.bah.append((self.maxBalls,self.ts))
        else:
            self.rah.append((self.maxRadius,self.ts))

    def change_arg(self,newArg):
        assert newArg[0] in {'b','r'}, "invalid argument"
        if newArg[0] == 'b':
            self.maxBalls = newArg[1]
        else:
            self.maxRadius = newArg[1]
        return


'''
file determines variance measures
'''
import numpy as np

# INPUT YOUR NUMBER HERE.
rng = np.random.default_rng()

def random_matrix_by_normaldist_values(n, mInfo):
    '''
    generates a random matrix with each column containing randomly-generated values corresponding to its (mean, std. dev.).
    '''

    q = np.zeros((n, len(mInfo)))
    for i in range(n):
        q[i] = one_random_sample_by_normaldist_values(mInfo)
    return q

'''
'''
def one_random_sample_by_normaldist_values(MInfo):

    s = np.zeros(len(MInfo),)
    for (i,m) in enumerate(MInfo):
        s[i] = rng.normal(m[0], m[1])
    return s
# violation_handler



from morebs2 import point_weight_function
import unittest
import numpy as np

'''
python -m morebs2.tests.point_weight_function_test  
'''
class TestPointWeightFunction(unittest.TestCase):

    '''
    '''
    def test__angle_between_two_points_clockwise(self):

        p = (55.0,65.0)

        # deg 0
        p1 = (56.0,65.0)
        # deg 90
        p2 = (55.0, 66.0)
        # deg 180
        p3 = (54.0, 65.0)
        # deg 270
        p4 = (55.0,64.0)

        assert point_weight_function.angle_between_two_points_clockwise(p,p1) == 0.0, "wrong angle for 0.0"
        assert point_weight_function.angle_between_two_points_clockwise(p,p2) == 90.0, "wrong angle for 00.0"
        assert point_weight_function.angle_between_two_points_clockwise(p,p3) == 180.0, "wrong angle for 180.0"
        assert point_weight_function.angle_between_two_points_clockwise(p,p4) == 270.0, "wrong angle for 270.0"

    '''
    uniform weights
    '''
    def test__similarity_measure__type43___uw_(self):

        bounds = np.array([[-70.0,60],[60,-70.0], [0.0,60], [-30.0,25]])
        weights = np.asarray(np.ones(4),dtype='float')

        # at opposing ends
        v1 = np.copy(bounds[:,0])
        v2 = np.copy(bounds[:,1])
        sm = point_weight_function.similarity_measure__type43(v1,v2, bounds, weights)
        assert sm == 1.0, "invalid 1"

        # test for commutativity
        sm = point_weight_function.similarity_measure__type43(v2,v1, bounds, weights)
        assert sm == 1.0, "invalid 2"

        # at same point: center
        v3 = np.copy(v1)
        v1 = v1 + (v2 - v1) / 2
        v2 = np.copy(v1)
        sm = point_weight_function.similarity_measure__type43(v1,v2,bounds,weights)
        assert sm == 0.0, "invalid 3"

        # at same point: end
        v1 = np.copy(v3)
        v2 = np.copy(v3)
        sm = point_weight_function.similarity_measure__type43(v1,v2,bounds,weights)
        assert sm == 0.0, "invalid 4"

        # at opposing ends +/- 1/2
        v1 = np.copy(bounds[:,0])
        v2 = np.copy(bounds[:,1])
        diff = bounds[:,1] - bounds[:,0]
        v1 += (diff / 4.0)
        v2 -= (diff / 4.0)
        sm = point_weight_function.similarity_measure__type43(v1,v2,bounds,weights)
        assert sm == 0.5, "invalid 5"

    # TODO: more testing needed.
    def test__similarity_measure__type43___nuw_(self):

        #
        bounds = np.array([[-70.0,60],[60,-70.0], [0.0,60], [-30.0,25]])
        weights = np.array([0.5,0.4,0.0, 0.25])

        # at opposing ends
        v1 = np.copy(bounds[:,0])
        v2 = np.copy(bounds[:,1])

        sm = point_weight_function.similarity_measure__type43(v1,v2,bounds,weights)
        weights = np.asarray(np.ones(4),dtype='float')
        sm1 = point_weight_function.similarity_measure__type43(v1,v2,bounds,weights)
        assert sm < sm1, "wrong relation for measure"

    def test__similarity_measure__type44___uw_(self):

        bounds = np.array([[-70.0,60],[60,-70.0], [0.0,60], [-30.0,25]])
        weights = np.asarray(np.ones(8),dtype='float')
        weights = weights.reshape((4,2))
        weights[:,1] = weights[:,1] * -1
        ##print("WEIGHTS ", weights)

        # case : `basically` at left endpoint
        v1 = np.array([-70.0,60,0.5, -30.0])
        v2 = np.copy(bounds[:,1])
        sm = point_weight_function.similarity_measure__type44(v1,v2,bounds,weights)
        ##print("SM ",sm)

        # case: `basically` at right endpoint
        v1 = np.copy(bounds[:,0])
        v2 = np.array([60.0,-69.5,60.0,25.0])
        sm1 = point_weight_function.similarity_measure__type44(v1,v2,bounds,weights)
        ##print("SM ",sm1)
        assert sm1 < sm, "[0] invalid sim. measure"

        sm1_ = point_weight_function.similarity_measure__type43(v1,v2,bounds,weights[:,0])
        assert sm1_ == sm1, "[1] invalid sim. measure"

        ###assert abs(sm1 - 1.0) == abs(sm - 1.0), "[2] incorrect relation"

        # case: v2 is at 1/2 + 1/4
        diff = bounds[:,1] - bounds[:,0]
        v1 = np.copy(bounds[:,0])
        v2 = np.copy(bounds[:,1]) - (diff * 3/8.0)
        sm2 = point_weight_function.similarity_measure__type44(v1,v2,bounds,weights)
        ##print("SM ",sm2)

        # case: v2 is at 1/2
        v2 = np.copy(bounds[:,1]) - (0.5 * diff)
        sm3 = point_weight_function.similarity_measure__type44(v1,v2,bounds,weights)
        ##print("SM ",sm3)
        assert sm2 > sm3, "[2] slant to right"

        # case: switch v1 and v2
        v1_ = np.copy(bounds[:,1])
        v2_ = np.copy(bounds[:,0]) + (0.5 * diff)
        sm3_ = point_weight_function.similarity_measure__type44(v1_,v2_,bounds,weights)
        ##print("@SM ", sm3_)
        sm3__ = point_weight_function.similarity_measure__type44(v2_,v1_,bounds,weights)
        ##print("@SM ",sm3_)
        assert sm3_ == sm3__, "[3] measure must be commutative"

        # case: v2 is at 3/8
        v2 = np.copy(bounds[:,0]) + (3.0/8.0 * diff)
        sm4 = point_weight_function.similarity_measure__type44(v1,v2,bounds,weights)
        ##print("SM ",sm4)
        assert sm3 < sm4

        # case: v2 is at 2^(-17)
        v2 = np.copy(bounds[:,0]) + (1.0/131072.0 * diff)
        sm5 = point_weight_function.similarity_measure__type44(v1,v2,bounds,weights)
        assert sm5 == 0.5, "incorrect sim. measure"

        v1 = np.copy(bounds[:,1])
        v2 = np.copy(bounds[:,1]) - (1.0/131072.0 * diff)
        sm6 = point_weight_function.similarity_measure__type44(v1,v2,bounds,weights)
        assert sm6 == 1.0 + (1.0 - sm5), ""

    def test___equal_matrices(self):
        m = np.arange(16).reshape((4,4))
        n = np.arange(16).reshape((4,4))
        assert point_weight_function.equal_iterables(m,n), "basic equality test failed"

########################### start: some tests on `matrix_methods`

    def test_intersection_of_bounds___1(self):

        # case 1: 1/2 of b2 in b1, b2 on right
        b1 = np.array([[0,40],[-20,59],[-299,-174],[60,112],[-20,65]])

            # make b2 by taking half
        b2 = []
        for i in range(b1.shape[0]):
            diff = (b1[i,1] - b1[i,0]) / 2
            s = b1[i,0] + diff
            e = s + diff
            b2.append([s,e])

        b2 = np.array(b2)
        intersection = point_weight_function.intersection_of_bounds(b1,b2)
        trueIntersection = [[  20.,    40. ],\
                            [  19.5,   59. ],\
                            [-236.5, -174. ],\
                            [  86.,   112. ],\
                            [  22.5,   65. ]]

        assert point_weight_function.equal_iterables(intersection,trueIntersection), "bounds intersection incorrect"

        # case 2: end bound of 1 is startbound of second
        b31 = b1[:,1] + 20.0
        b30 = np.copy(b1[:,1])
        b3 = np.vstack((b30,b31)).T

        intersection = point_weight_function.intersection_of_bounds(b1,b3)
        assert type(intersection) == type(None), "[3] invalid intersection"

        # case 3: no intersection at all
        b40 = np.copy(b3[:,1])
        b41 = b40 + 50.0
        b4 = np.vstack((b40,b41)).T

        intersection = point_weight_function.intersection_of_bounds(b1, b4)
        assert type(intersection) == type(None), "[4] invalid intersection {}".format(intersection)

    def test_intersection_of_bounds___2(self):
        # case 1: b2 in b1, and b1 bounds are not sorted
        b1 = np.array([[40,0],[-20,59],[-174,-299],[60,112],[-20,65]])
        b2 = np.array([[0,10], [10,-10],[-299,-250],[90,110],[10,-15]])
        intersection = point_weight_function.intersection_of_bounds(b1,b2)

        trueIntersection = [[   0.,   10.],\
                            [ -10.,   10.],\
                            [-299., -250.],\
                            [  90.,  110.],\
                            [ -15.,   10.]]
        assert point_weight_function.equal_iterables(intersection,trueIntersection)

    def test_euclidean_point_distance(self):
        p1 = np.array([4,15,20,7])
        p2 = np.array([14,20,21,12])
        q = point_weight_function.euclidean_point_distance(p1,p2)
        assert abs(q - 12.288) < 10 ** -3

if __name__ == '__main__':
    unittest.main()



from .matrix_methods import * 
from copy import deepcopy

"""
given: a sequence S, either an np.array or a list 
output: sequence S' corresponding to `S`. Every 
    element of S' is a pair of the form 
    [0] value v of S at the j'th index 
    [1] number of contiguous indices, from the 
        j'th index in increasing order, with 
        values that equal v. 

EX: 
S = <0,0,1,2,1,2,3,3,4,4,4>
S' = <(0,2),(1,1),(2,1),(1,1),(2,1),(3,2),(4,2)> 
"""
def contiguous_repr__sequence(S):
    assert is_vector(S) or type(S) == list
    assert len(S) > 0

    q = [[S[0],1]]
    ref = q[-1]

    for i in range(1,len(S)):
        if S[i] == ref[0]:
            ref[1] += 1
        else:
            ref_ = [S[i],1]
            q.append(ref_) 
            ref = ref_
    return q

def contiguous_repr_size(S,size_type):
    return -1 

"""
V is a vector of (value,frequency) pairs 
assumed to be sorted in ascending or descending 
order. Function finds the sequence of all 
(value,frequency) pairs that tie for a place 
in the ranking of frequency. 
"""
def valuefreq_pair_vector__nth_place(V,i=0):

    j = None

    ip = 0
    ref = V[0]
    seq = []
    for i_ in range(1,len(V)):
        if ip == i:
            if ref[1] != V[i_][1]:
                break
            else: 
                seq.append(V[i_][0]) 
            continue
    
        if ref[1] != V[i_][1]:
            ip += 1 
            ref = V[i_] 
    return seq

def valuefreq_pair_vector_to_tie_partition(V):

    j = None
    ip = 0
    ref = V[0]
    seqs = []
    seq = []
    for i_ in range(1,len(V)):
        if ref[1] != V[i_][1]:
            ip += 1 
            ref = V[i_] 
            seqs.append(seq) 
            seq = [V[i_][0]]
        else: 
            seq.append(V[i_][0]) 
    return seqs 

"""
most common contiguous subsequence search 
"""
class MCSSearch:

    def __init__(self,L,cast_type=int,is_bfs:bool=True):  
        assert type(L) == list or is_vector(L) 
        assert type(is_bfs) == bool
        self.l = np.array(L)  
        self.cast_type = cast_type 
        self.is_bfs = is_bfs 

        self.preproc()
        # stringized subsequence of L --> index list of occurrence 
        self.subseq_occurrences = defaultdict(list) 
        self.key_queue = [] 
        self.key_cache = []
        return
    
    def preproc(self):
        self.d2index = defaultdict(list) 
        for (i,l_) in enumerate(self.l):
            self.d2index[l_].append(i) 
    
    def most_frequent_(self):
        q = sorted([(k,len(v)) for k,v in self.d2index.items()],key=lambda x:x[1],reverse=True) 

        q_ = q.pop(0)   
        s = set() 
        s |= {q_[0]} 
        while len(q) > 0:
            q2_ = q.pop(0) 
            if q2_[1] != q_[1]:
                break 
            s |= {q2_[0]} 
        return s 
    
    """
    post-search main method 
    """
    def mcs(self):
        x = [(k,len(v)) for k,v in self.subseq_occurrences.items()] 
        x = sorted(x,key=lambda x:x[1])
        q = x.pop(-1)
        s = [string_to_vector(q[0],castFunc=self.cast_type)]
        while len(x) > 0:
            q_ = x.pop(-1)
            if q_[1] == q[1]:
                s_ = string_to_vector(q_[0],castFunc=self.cast_type)
                s.append(s_)
            else: 
                break 
        return s 

    def mcs_nth(self,i=0): 
        x = [(k,len(v)) for k,v in self.subseq_occurrences.items()] 
        x = sorted(x,key=lambda x_:x_[1],reverse=True)
        return valuefreq_pair_vector__nth_place(x,i)
    
    def init_search(self):

        self.subseq_occurrences.clear() 
        self.key_queue.clear()
        self.key_cache.clear() 

        #ss = sorted(self.most_frequent())
        ss = sorted(set(np.unique(self.l)))

        for ss_ in ss: 
            v = self.d2index[ss_] 
            ssv = vector_to_string([ss_],castFunc=self.cast_type)
            self.subseq_occurrences[ssv] = v
            self.key_queue.append(ssv)  
        return
    
    def __next__(self):
        if len(self.key_queue) == 0: 
            return False 
        
        x = self.key_queue.pop(0)
        q = self.extend_subseq(x) 

        self.key_cache.append(x) 

        # sort q by frequency
        q_ = []
        for q2 in q:
            v = self.subseq_occurrences[q2] 
            q_.append((q2,v)) 
        q_ = sorted(q_,key=lambda x:x[1],reverse=True)  
        q_ = [q2[0] for q2 in q_] 

        if self.is_bfs:
            self.key_queue.extend(q_)
        else:
            while len(q_) > 0:
                self.key_queue.insert(0,q_.pop(-1)) 
        return True 

    def extend_subseq(self,subseq_str): 
        q = self.subseq_occurrences[subseq_str] 
        subseq_base = string_to_vector(subseq_str,castFunc=self.cast_type)
        new_subseq_ = set() 
        for q_ in q:
            # get the next index
            i = len(subseq_base) + q_ 

            if i >= len(self.l): continue

            subseq = deepcopy(subseq_base) 
            subseq = np.append(subseq,self.l[i])

            subseq_str_ = vector_to_string(subseq,castFunc=self.cast_type) 
            self.subseq_occurrences[subseq_str_].append(q_)  
            new_subseq_ |= {subseq_str_}
        return new_subseq_ 
    
    """
    pre-search main method 
    """
    def search(self):
        self.init_search() 

        stat = True 
        while stat: 
            stat = self.__next__()
        return 


from morebs2.deline import *

class DPointAnalysis:
    '''
    provides a description of the complexity of the dataset
    of two-dimensional labelled points that :class:`DLineateMC`
    delineates. 
    '''

    def __init__(self):
        return

class DLineateMetric:
    '''
    :attribute label:
    :attribute posMiss: points found inside delineation with a different label
    :attribute negMiss: points found outside delineation with the same label 
    '''

    def __init__(self,label,counter):
        self.label = label
        self.count = counter[label]
        self.posMiss = np.sum(np.array(list(counter.values())))\
            - counter[label]
        self.negMiss = counter[-1]

class DLineateMC:
    '''
    NOTE: inefficiencies exist. 
    '''

    def __init__(self,xyl,dmethod="nodup"):
        self.xyl = xyl
        self.dmethod = dmethod
        self.xyl_unproc = deepcopy(self.xyl)
        self.d = None
        self.dds = []
        self.last_process = [None,None]
        self.num_labels_rem = None
        self.didn = "0"
        self.bilabel = None # [delineator id, label]

    def delineate_one(self,clockwise = True):
        # case: done
        if self.num_labels_rem == 1:
            return True

        self.declare_DLineate22(clockwise)
        l = self.d.idn
        c = self.delineate_remainder(clockwise)

        # case: only two remaining
        if self.num_labels_rem == 2:
            q = list(c.values())
            q_ = None
            for q2 in q: 
                if q2 != self.d.label: 
                    q_ = q2
                    break
            self.bilabel = [l,q_] 
        return False

    def delineate_remainder(self,clockwise=True):
        '''
        delineates the points in `xyl`

        return := bool to further delineate on the points in `d`. 
        '''

        i,c,q = self.d.full_process(False)
        print("I")
        print(i)
        print("C")
        print(c)
        print("Q")
        print(q)

        self.dds.append(self.d)
        # case: perfect delineation
        if c[self.d.label] == c[-1] + c[self.d.label]:
            #self.xyl_unproc = np.delete(self.xyl_unproc,i,0)
            return c 

        self.delineate_delineation(c,i,clockwise)
        #self.xyl_unproc = np.delete(self.xyl_unproc,i,0)
        return c

    def delineate_delineation(self,c,indices,clockwise):
        '''
        '''
        assert type(self.d) != type(None), "no delineation to delineate"

        # declare a disposable `DLineate22` to get labelCounts
        d = DLineate22(deepcopy(self.xyl_unproc[indices]),
            clockwise,"nojag",self.didn)
        d.label_counts()
        lc = deepcopy(d.lc)

        for c_ in c:
            if c_ == self.d.label:
                continue
            
            d = self.declare_sub_DLineate22(c_,indices,lc)
            i2,c2,q2 = d.full_process(False)

            # get increase and decrease score
            inc = q2
            dec = sum(c2.values()) - q2
            inc -= dec

            # case: improvement, add to solution
            if inc > 0:
                self.dds.append(d)
            else:
            # case: no improvement, do not add
                self.decrement_didn()

        return

    def declare_sub_DLineate22(self,targetLabel,indices,labelCounts):
        d = DLineate22(deepcopy(self.xyl_unproc[indices]),
            clockwise,self.dmethod,self.didn)
        self.increment_didn()
        d.lc = deepcopy(labelCounts)
        d.set_target_label(targetLabel)
        d.sort_data()
        return d

    def declare_DLineate22(self,clockwise):
        '''
        '''

        # remove processed points from
        if type(self.d) != type(None):
            self.xyl_unproc = deepcopy(self.d.xyl)

        self.d = DLineate22(deepcopy(self.xyl_unproc),clockwise,"nojag",self.didn)
        self.increment_didn()
        self.d.preprocess()


        # get the number of labels remaining
        self.num_labels_rem = len(self.d.lc)
        return

    def increment_didn(self):
        self.didn = str(int(self.didn) + 1)

    def decrement_didn(self):
        self.didn = str(int(self.didn) - 1)

def test_dataset__DLineateMC_2():
    '''
    points of different labels spatially intersect; 5 labels. 
    '''
    return -1

def test_dataset__DLineateMC_3():
    '''
    points of different labels spatially intersect (entirely); 4 labels. 
    '''
    return -1

########


"""
file contains code for distributions
"""

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from math import sqrt,pi,e,log

import random

random.seed(100)

############################# start: plotting functions

def basic_2d_lineplot(x,y):
    fig, ax = plt.subplots()  # Create a figure containing a single axes.
    ax.plot(x, y)
    plt.show()


def basic_2d_scatterplot(x,y,c=None):
    plt.scatter(x, y,c=c,s=1)#, s=area, c=colors, alpha=0.5)
    plt.show()

def generate_2d_data_from_function(f,xRange,xInc,additionalArgs = ()):
    """generates a sequence of (x,y) pairs such that every x-value has exactly one y-value and the x-values are evenly distributed across the x-range.
    """
    
    x = []
    y = []
    q = xRange[0]
    while q <= xRange[1]:
        x.append(q)
        y.append(f(q,*additionalArgs))
        q = q + xInc
    return x,y

def generate_2d_data_from_function__multipleY(f,xRange,xInc,additionalArgs = ()):
    """generates a sequence of (x,y) pairs such that every x-value can have any integer n >= 0 of y-values according to arguments. 
    """

    x = []
    y = []
    q = xRange[0]
    while q <= xRange[1]:
        ys = f(q,*additionalArgs) 
        if len(ys) != 0:
            x.extend([q] * len(ys))
            y.extend(ys)
        q = q + xInc
    return x,y

def function_to_2dplot(g,f,xRange,xInc,additionalArgs=(),mode="scatter"):
    assert mode in ["line","scatter"]
    x,y = g(f,xRange,xInc,additionalArgs)
    if mode == "line":
        basic_2d_lineplot(x,y)
    else:
        basic_2d_scatterplot(x,y)
    return

############################# end: plotting functions

############################# start: distribution functions

def normal_distribution_function(x,mean,stddev):
    m = 1.0 / sqrt(2. * pi)
    exp = -0.5 * ((x - mean) / (stddev**2.))**2
    return m * e ** exp
    
def holtzmann_distribution_function():
    return -1

def poisson_distribution_function():
    return -1

def multiple_y_function_stdrand(x,yRange,multipleRange):
    """generates multiple y-values for the x-value based on python standard random. The x-value has no bearing on the y-values generated.
    """

    assert yRange[1] >= yRange[0], "invalid range for y-values"
    assert multipleRange[1] >= multipleRange[0], "invalid range for multiples"

    # determine the number of points
    p1 = multipleRange[0] + int((multipleRange[1] - multipleRange[0]) * random.random())
    l = []
    for i in range(p1):
        l.append(yRange[0] + (yRange[1] - yRange[0]) * random.random())
    return l

def multiple_y_function_x_effector(x,f,multipleRange):
    """generates variable # of y-values for the 
    x-value in which the x-value has an 
    effect on the y-values generated.
    """

    p1 = multipleRange[0] + int((multipleRange[1] - multipleRange[0]) * random.random())
    l = []
    for i in range(1,p1 + 1):
        l.append(f(x,i))
    print("L: ", len(l))
    return l

def x_effector_function_type_modulo(xyiOp,yRange):
    """outputs a function 
                `f(x,i) -> y`
        using a function `xyiOp(x,yRange,i)`. 

    :param xyiOp: function
    :param yRange: range of the y-values 
    """

    def q(x,i):
        return xyiOp(x,yRange,i)
    return q

def xyi_op_type_1(x,yRange,i):
    r = yRange[1] - yRange[0]
    x2 = (x * i * 10) % r
    return yRange[0] + x2

def log_function(x,m=e):
    return log(x) / log(m)

######### test function

### gaussian dist.
#function_to_2dplot(generate_2d_data_from_function,normal_distribution_function,[-5.,5.],0.1,(0.,5.))

### multiple y function 1
#f = x_effector_function_type_modulo(xyi_op_type_1,[200.,600.])#[-112.5,400.3])
#function_to_2dplot(generate_2d_data_from_function__multipleY,\
#    multiple_y_function_x_effector,[-20.,40.],0.25,additionalArgs=(f,[0,2]),mode="scatter")

### multiple y function 2
#function_to_2dplot(generate_2d_data_from_function__multipleY,\
#    multiple_y_function_stdrand,[-20.,40.],0.25,additionalArgs=([100.,170.],[1,6]),mode="scatter")

############################# end: distribution functions


from .ball_operator_test_cases import *

def test__BallOperator__run_subball_split():
    print("SUB-BALL SPLIT TEST")
    balls = sample_balls_1()

    # choose 3 random
    ballIndices = [0,3,5]

    b = BallOperator(balls[ballIndices[0]])

    print("initial")
    print(balls[ballIndices[0]])
    print()

    r = balls[ballIndices[0]].radius / 1.1
    b.run_subball_split((r,"literal"),"minimal")

    print("after split")
    print("number of balls ", len(b.subballs))
    for b_ in b.subballs:
        print(b_)
        print()

    print("average subball radius")
    radius = np.mean([b_.radius for b_ in b.subballs])
    print(radius)
    print("average subball data size")
    ds = np.mean([b_.data.newData.shape[0] for b_ in b.subballs])
    print(ds)
    print()

    print("base ball")
    print(b.ball)

    # comment the line below to see information on ball that was split
    return -1

    fp = BallClumpDataGenerator.ballclump_frame_filepath(ballClumpDataFilePath1)
    fobj = pickle_obj_from_file(fp)

    print("--- Ball information")
    print("cluster centers")
    print(fobj[3][0])
    print()

    print("cluster radii")
    print(fobj[4][0])

    print("cluster densities")
    print(fobj[5][0])
    print()
    print("-------------------------------")
    return



'''
this file tests the subset of
methods found in `matrix methods`
that are for bounds-related calculations
'''
from morebs2 import matrix_methods
import unittest
import numpy as np

'''
python -m morebs2.tests.matrix_methods__bounds__test  
'''
class TestNumericalGeneratorClass(unittest.TestCase):

    def test__point_in_improper_bounds(self):

        pb = np.array([[0.4,12],[-30,40],[70,112],[10,14]])
        b = np.array([[10,2],[9,-28], [93,72],[13.5,11.0]])
        q = matrix_methods.split_improper_bound(pb,b)
        p1 = np.array([3,-25,77,11.7]) # not
        p2 = np.array([11,39,102,10.1])

        x = matrix_methods.point_in_improper_bounds(pb,b,p1)
        assert not matrix_methods.point_in_improper_bounds(pb,b,p1), "incorrect case 1"
        assert matrix_methods.point_in_improper_bounds(pb,b,p2), "incorrect case 2"
        return

    def test__point_on_improper_bounds_by_ratio_vector(self):

        pb = np.array([[0.4,12],[-30,40],[70,112],[10,14]])
        b = np.array([[10,2],[9,-28], [93,72],[13.5,11.0]])

        # case 1:
        v = np.array([0.5,0.5,0.5,0.5])
        q = matrix_methods.point_on_improper_bounds_by_ratio_vector(pb,b,v)
        assert matrix_methods.point_in_improper_bounds(pb,b,q), "incorrect case 1.1"
        assert matrix_methods.equal_iterables(q,np.array([11.8,25.5,103.5,10.25])), "incorrect case 1.2"

        # case 2:
        v = np.array([0.9,0.9,0.8,0.1])
        q = matrix_methods.point_on_improper_bounds_by_ratio_vector(pb,b,v)
        assert matrix_methods.point_in_improper_bounds(pb,b,q), "incorrect case 2.1"
        assert matrix_methods.equal_iterables(q,np.array([1.64,38.7,109.8,13.65])), "incorrect case 2.2"
        return

    def test__vector_ratio_improper(self):

        pb = np.array([[0.4,12],[-30,40],[70,112],[10,14]])
        b = np.array([[10,2],[9,-28], [93,72],[13.5,11.0]])

        p1 = np.array([11.8,25.5,103.5,10.25])
        p2 = np.array([1.64,38.7,109.8,13.65])

        vr1 = matrix_methods.vector_ratio_improper(pb,b,p1)
        assert matrix_methods.equal_iterables(vr1,np.array([0.5,0.5,0.5,0.5])), "incorrect case 1.1"

        vr2 = matrix_methods.vector_ratio_improper(pb,b,p2)
        assert matrix_methods.equal_iterables(vr2,np.array([0.9,0.9,0.8,0.1])), "incorrect case 1.2"


    def test__point_difference_of_improper_bounds(self):

        pb = np.array([[0,1.00],\
                        [0,1.00],\
                        [0,1.00],\
                        [0,1.00]])

        # one index is improper
        bp1 = np.array([[0.5,0.75],\
                        [0.25,0.50],\
                        [0.9,0.1],\
                        [0,0.27]])

        # two indices are improper
        bp2 = np.array([[0.5,0.75],\
                        [0.5,0.15],\
                        [0.9,0.1],\
                        [0,0.27]])

        # all indices are improper
        bp3 = np.array([[0.5,0.17],\
                        [0.5,0.15],\
                        [0.9,0.1],\
                        [0.72,0.3]])

        q = matrix_methods.point_difference_of_improper_bounds(bp1,pb)
        s1 = np.array([0.25,0.25,0.2,0.27])
        assert matrix_methods.equal_iterables(q,s1)

        q = matrix_methods.point_difference_of_improper_bounds(bp2,pb)
        s2 = np.array([0.25,0.65,0.2,0.27])
        assert matrix_methods.equal_iterables(q,s2)

        q = matrix_methods.point_difference_of_improper_bounds(bp3,pb)
        s3 = np.array([0.67,0.65,0.2,0.58])
        assert matrix_methods.equal_iterables(q,s3)
        return

    def test__point_on_improper_bounds_by_ratio_vector(self):

        pb = np.array([[0,1.00],\
                        [0,1.00],\
                        [0,1.00],\
                        [0,1.00]])

        # one index is improper
        bp1 = np.array([[0.5,0.75],\
                        [0.25,0.50],\
                        [0.9,0.1],\
                        [0,0.27]])

        # half
        rv = np.array([0.5,0.5,0.5,0.5])
        p1 = matrix_methods.point_on_improper_bounds_by_ratio_vector(pb,bp1,rv)
        assert matrix_methods.equal_iterables(p1,[0.625,0.375,0.,0.135])

        # max dominant
        rv = np.array([0.9,0.9,0.9,0.9])
        p1 = matrix_methods.point_on_improper_bounds_by_ratio_vector(pb,bp1,rv)
        assert matrix_methods.equal_iterables(p1,[0.725,0.475,0.08,0.243])

        # min dominant
        rv = np.array([0.1,0.1,0.1,0.1])
        p1 = matrix_methods.point_on_improper_bounds_by_ratio_vector(pb,bp1,rv)
        assert matrix_methods.equal_iterables(p1,[0.525,0.275,0.92,0.027])

    def test__submatrix__2d(self): 

        x = np.arange(1,21)
        x = x.reshape((4,5))
        q0 = matrix_methods.submatrix__2d(x,(1,2),"L+U")
        assert np.all(q0 == np.array(\
        [[ 6,  7,  8],\
        [11, 12, 13],\
        [16, 17, 18]]))

        q1 = matrix_methods.submatrix__2d(x,(1,2),"L+L")
        assert np.all(q1 == np.array([\
        [ 8,  9, 10],\
        [13, 14, 15],\
        [18, 19, 20]]))

        q2 = matrix_methods.submatrix__2d(x,(1,2),"R+L")
        assert np.all(q2 == np.array([\
        [ 3,  4,  5],\
        [ 8,  9, 10]])) 

        q3 = matrix_methods.submatrix__2d(x,(1,2),"R+U")
        assert np.all(q3 == np.array([\
        [1, 2, 3],\
        [6, 7, 8]])) 

    def test__index_range_of_subvec__case1(self):

        v = [0,10,13,14,17,19,20,21,21,21,0,20,21,10,14,17] 
        sv = [17,19,20,21] 
        ir1 = matrix_methods.index_range_of_subvec(v,sv,is_contiguous=True)
        assert v[ir1[0]:ir1[1]] == sv 

        sv1 = [0,10,17,21]
        ir1 = matrix_methods.index_range_of_subvec(v,sv1,is_contiguous=False)
        assert ir1 == (0,7)

        sv2 = [10,14,19,14,17] 
        ir2 = matrix_methods.index_range_of_subvec(v,sv2,is_contiguous=False)
        assert ir2 == (1,15)

        sv3 = [10,14,19,14,17100]  
        ir3 = matrix_methods.index_range_of_subvec(v,sv3,is_contiguous=False)
        assert type(ir3) == type(None)


if __name__ == '__main__':
    unittest.main()



'''
file practices with techniques to navigate a ball.

ATTENTION:
methods will get transferred to test after completion
'''
from .ball_operator_test_cases import *

#######-------------------------------------------------

def test__BallOperator__nav_basis():
    p = np.array([13,41,50,71,-10])
    b = BallOperator.nav_basis(p)
    sol = np.array([[13,41,50,71,-10],\
            [-10,13,41,50,71],\
            [71,-10,13,41,50],\
            [50,71,-10,13,41],\
            [41,50,71,-10,13]])
    assert equal_iterables(b,sol), "incorrect navigational basis"
    return

def test__BallOperator__set_navigation__AND__basis_division():
    ball = sample_ball_x()
    b = BallOperator(ball)

    # test: set_navigation
    p = np.copy(ball.radiusDelta[0])
    b.set_navigation(p)
    assert equal_iterables(b.location - ball.center,-(b.counterLocation - ball.center))

    # test:
    multiplier = np.array([-1,1,1,-1])
    point = np.array([-39,-1.75,32.5,-7])
    x = BallOperator.basis_division(b.location,b.counterLocation,multiplier)

    assert equal_iterables(x[0],point), "incorrect basis division"
    return

def test__BallOperator__hop_in_division():

    p = np.array([13,41,50,71,-10])
    b = BallOperator.nav_basis(p)

    center = np.array([0.0,0.0,0.0,0.0,0.0])

    hid = BallOperator.hop_in_division(b,0.0,center)
    sol = np.copy(center)
    assert equal_iterables(hid,sol), "incorrect case 1"

    hid = BallOperator.hop_in_division(b,0.25,center)
    d = 1/4 * (np.array([-10,13,41,50,71]) - np.array([13,41,50,71,-10]))
    sol = np.array([13,41,50,71,-10]) + d
    assert equal_iterables(hid,sol), "incorrect case 2"

    hid = BallOperator.hop_in_division(b,0.1,center)
    sol = np.array([6.5,20.5,25.,35.5,-5.])
    assert equal_iterables(hid,sol), "incorrect case 3"

    hid = BallOperator.hop_in_division(b,1.0,center)
    sol = np.array([41,50,71,-10,13])
    assert equal_iterables(hid,sol), "incorrect case 4"

    hid = BallOperator.hop_in_division(b,0.8,center)
    sol = np.array([50,71,-10,13,41])
    assert equal_iterables(hid,sol), "incorrect case 5"
    return



# some more functions for class<RChainHead>. 
import numpy as np

def vector_summation_modulo(m,v):
    '''
    multiplies all values of `v` by 10^5 to convert them to integers. Then 
    adds all values of `v` into a sum and modulo `m` that sum.

    :param v:
    :param m: 
    :return: non-negative integer less than `m`
    :rtype: int
    '''

    d = decimal_order_of_vector(v)
    v = np.array([int(v_) for v_ in v * 10**d])
    x = np.sum(v)
    return x % m

def decimal_order_of_vector(v):
    '''
    maximum decimal places of values in vector
    '''

    return max([decimal_places_of_float(v_) for v_ in v])

def decimal_places_of_float(f,maxPlaces = 5):
    for i in range(maxPlaces):
        f_ = f * 10 ** i
        if int(f_) == f_: return i
    return maxPlaces

def update_modulo_function(a,m2):

    def update_f(m):
        return (m + a) % m2

    return update_f



# NEXT: subvector selector
# NEXT: add verbose mode for processing each vector in NSDataInstruction
# NEXT: add verbose mode for RChainHead


from PIL import Image
from pathlib import PurePath
from bs4 import UnicodeDammit
from .matrix_methods import *

STREAM_BLOCK_SIZE_RANGE = (24, 633)
DEFAULT_STREAM_BLOCK_SIZE = 63

class MessageStreamer:

    acceptableFileExtensionForText = [".txt"]
    acceptableFileExtensionForImage = [".jpg"]
    acceptableEncodings = ["ascii", "utf_8","utf-8"]

    def __init__(self, filePath, streamBlockSize = DEFAULT_STREAM_BLOCK_SIZE,readMode = "rb"):
        """
        class is to "stream" data from .a file source, (image = jpg file)|(text = 'literal'|'csv'), set `readMode` to "r" for reading csv files
        """
        assert readMode in ["rb","r"]
        self.filePath = PurePath(filePath)
        self.rm = readMode
        self.file_to_msg_alphabet()
        self.openedFile = None
        self.open_file()
        assert streamBlockSize >= STREAM_BLOCK_SIZE_RANGE[0] and streamBlockSize <= STREAM_BLOCK_SIZE_RANGE[1], "stream block size inaccurate"
        self.streamBlockSize = streamBlockSize
        self.blockData = []

    def file_to_msg_alphabet(self):
        if self.filePath.suffix in MessageStreamer.acceptableFileExtensionForImage:
            self.messageAlphabet = MessageAlphabet.pixelColor
            self.mav = "pic"
        elif self.filePath.suffix in MessageStreamer.acceptableFileExtensionForText:
            self.messageAlphabet = MessageAlphabet.alphanumeric
            self.mav = "tex"
        else:
            raise ValueError("file {} is invalid message".format(self.filePath))

    def open_file(self):
        if self.mav == "pic":
            self.openedFile = Image.open(str(self.filePath))
            self.loadedPic = self.openedFile.load() # CAUTION: loads entire image into memory
            self.streamIndex = (0,0)
        else:
            self.check_valid_text_file_encoding()
            self.openedFile = open(str(self.filePath),self.rm)
            self.streamIndex = 0

        return

    def check_valid_text_file_encoding(self):
        if self.mav != "tex": return

        with open(str(self.filePath), 'rb') as fi:
            content = fi.read(10 ** 5)
            suggestion = UnicodeDammit(content)
            self.textEncoding = suggestion.original_encoding
            assert self.textEncoding in MessageStreamer.acceptableEncodings, "text encoding {} is wrong".format(self.textEncoding)

    def stream_block(self):
        """
        :return: block of data, terminate stream, finished stream
        :rtype: list,bool,bool
        """

        self.blockData = []
        c = 0
        if self.mav == "tex":
            while c < self.streamBlockSize:

                # case: "rb"
                if self.rm == "rb":
                    b = self.openedFile.read(1)
                # case: "r"
                else:
                    b = self.openedFile.readline()
                    b = string_to_vector(b, castFunc = cr)

                if len(b) == 0:
                    break
                self.streamIndex += 1
                self.blockData.append(b)
                c += 1
        else:

            # x then y
            def next_index(i):
                i1,i2 = (i[0] + 1, i[1])

                if i1 >= self.openedFile.size[0]:
                    i1, i2 = 0, i2 + 1

                if i2 >= self.openedFile.size[1]:
                    return None, False

                return (i1,i2), True

            sz = self.openedFile.size
            while c < self.streamBlockSize:
                # read pixel
                v = self.loadedPic[self.streamIndex[0], self.streamIndex[1]]
                self.streamIndex, stat = next_index(self.streamIndex)
                if not stat: break
                self.blockData.append(v)
                c += 1
        return

    def stream(self):

        if type(self.openedFile) == type(None):
            return False
        self.stream_block()

        if len(self.blockData) == 0:
            self.end_stream()
            return False
        return True

    def end_stream(self):
        if type(self.openedFile) == type(None):
            return
        self.openedFile.close()
        self.openedFile = None

class MessageAlphabet:

    alphanumeric = [(32, 127)]
    pixelColor = [(0, 256), (0, 256), (0, 256)]



from morebs2 import numerical_space_data_generator,relevance_functions
import unittest
import numpy as np

'''
python -m morebs2.tests.numerical_space_data_generator_test  
'''

def sample_nsdi_1(q, filePath = "tests/s.txt", nr = None):
    """
    :param q: one of [relevance zoom,prg,np.ndarray]
    :param filePath: str, file path 
    :param nr: noise range,np.ndarray w/ length of 1 or 4. 
    """

    if type(q) == str:
        assert q in ['relevance zoom', 'prg']
    else:
        assert type(q) == np.ndarray, "333"

    bounds = np.array([[0,1.2],[-0.5,1.7],[-1.0,0.8],[-0.5,1.2]])
    sp = np.copy(bounds[:,0])
    columnOrder = None
    ssih = 4
    cv = 0.6
    bInf = (bounds,sp,columnOrder,ssih,(cv,))

    # make relevance function
    rch = relevance_functions.sample_rch_1_with_update(bounds, np.copy(bounds), ssih, cv)
    rm = (q, rch)

    ##sp = np.copy(bounds[:,1])
    modeia = 'w'

    q = numerical_space_data_generator.NSDataInstructions(bInf,rm,filePath,modeia,nr)
    return q

def sample_nsdi_2(q, filePath = "tests/s.txt", nr = None):
    """
    :param q: one of [relevance zoom,prg,np.ndarray]
    :param filePath: str, file path 
    :param nr: noise range,np.ndarray w/ length of 1 or 4. 
    """

    if type(q) == str:
        assert q in ['relevance zoom', 'prg']
    else:
        assert type(q) == np.ndarray, "333"

    bounds = np.array([[0,1.2],[-0.5,1.7],[-1.0,0.8],[-0.5,1.2]])
    sp = np.copy(bounds[:,0])
    columnOrder = None
    ssih = 7
    bInf = (bounds,sp,columnOrder,ssih,())

    # make relevance function
    rch = relevance_functions.sample_rch_2_with_update(bounds, np.copy(bounds))
    rm = (q, rch)
    modeia = 'w'

    q = numerical_space_data_generator.NSDataInstructions(bInf,rm,filePath,modeia,nr)
    return q


def sample_nsdi_3(wom=relevance_functions.idv,filePath = "ARGHONIAX"):
    assert wom == relevance_functions.idv or wom == relevance_functions.binary_labelling_scheme_1
    bounds = np.array([[0,33],[-3,42],[-3,15],[30,69],[20,200],[-10,370],[-100,100],[-250,100],[100,1000],[-400,250]])

    sp = np.copy(bounds[:,0])
    columnOrder = None
    ssih = 3
    bInf = (bounds,sp,columnOrder,ssih,())

    rch = relevance_functions.sample_rch_hop_counts_scheme_type_2(np.copy(bounds),np.copy(bounds),ssih,15,wom)
    rm = ("relevance zoom",sample_rch_blind_accept())

    nsdi = NSDataInstructions(bInf, rm,"ARGHONIAY.txt",'w',noiseRange = None,writeOutMode = rch)
    return nsdi 
    

class TestNSDataInstructionClass(unittest.TestCase):

    '''
    '''
    def test__sample_nsdi_11(self):
        print("Test 1 for: NS Data Instruction")
        q = sample_nsdi_1("relevance zoom",filePath="tests/s11.txt")
        q.make_rssi()

        c = 0
        while q.fp:
            q.next_batch_()
            c += 1
        print("case 1")
        print("# batches: ",c)
        q.batch_summary()

    def test__sample_nsdi_12(self):
        print("Test 2 for: NS Data Instruction")
        q = sample_nsdi_1('prg',"tests/s12.txt")
        q.make_rssi()

        c = 0
        while q.fp and c < 100:
            q.next_batch_()
            c += 1
        print("case 2")
        print("# batches: ",c)

        if q.fp:
            q.fp.close()

        q.batch_summary()

    def test__sample_nsdi_13(self):
        print("Test 3 for: NS Data Instruction")
        b1 = np.array([[0.9,1.2],[1.2,1.7],[-1.0,-0.5],[0.7,1.1]])
        bx = np.array([b1])

        q = sample_nsdi_1(bx,"tests/s13.txt")
        q.make_rssi()

        while q.fp:
            q.next_batch_()
        assert q.c == 3, "incorrect number of batches,got {} want {}".format(q.c,1)
        print("case 3")
        q.batch_summary()

    def test__sample_nsdi_14(self):
        print("Test 4 for: NS Data Instruction")
        nr = np.array([[0.01,0.07]])
        q = sample_nsdi_1('relevance zoom',"tests/s14.txt",nr)
        q.make_rssi()

        c = 0
        while q.fp and c < 100:
            q.next_batch_()
            c += 1
        if q.fp:
            q.fp.close()
        print("# batches: ",c)
        q.batch_summary()

    def test__sample_nsdi_21(self):
        print("Test 5 for: NS Data Instruction")
        q = sample_nsdi_2("relevance zoom",filePath="tests/s21.txt")
        q.make_rssi()
        while q.fp:
            q.next_batch_()
        return

    def test__sample_rch_blind_accept(self):
        bounds = np.array([[0,3],[-3,3],[-3,0],[3,6]])
        sp = np.copy(bounds[:,0])
        columnOrder = None
        ssih = 2
        bInf = (bounds,sp,columnOrder,ssih,())
        rm = ("relevance zoom",relevance_functions.sample_rch_blind_accept())
        nsdi = numerical_space_data_generator.NSDataInstructions(bInf, rm,"ARGHONIA.txt",'w',noiseRange = None,writeOutMode = "literal")
        nsdi.make_rssi()

        c = 0
        while nsdi.fp:
            nsdi.next_batch_()
            #nsdi.batch_summary()
            c += 1
        assert c == 11, "incorrect number of batches,got {} want {}".format(c,11) 

    def test__sample_rch_hop_counts_scheme_type_1(self):

        bounds = np.array([[0,3],[-3,3],[-3,0],[3,6]])
        sp = np.copy(bounds[:,0])
        columnOrder = None
        ssih = 3
        bInf = (bounds,sp,columnOrder,ssih,())
        rch = relevance_functions.sample_rch_hop_counts_scheme_type_1(np.copy(bounds),np.copy(bounds),ssih)
        rm = ("relevance zoom",relevance_functions.sample_rch_blind_accept())
        nsdi = numerical_space_data_generator.NSDataInstructions(bInf, rm,"ARGHONIA2.txt",'w',noiseRange = None,writeOutMode = rch)
        nsdi.make_rssi()

        while nsdi.fp:# and c < 1:
            nsdi.next_batch_()

if __name__ == '__main__':
    unittest.main()



from .matrix_methods import *
import math


class HopPattern:
    """
    class uses a 1-d number line as reference
    number line is finite between `minBound` and `maxBound`.
    """
    DEF_INCREMENT_RATIO = 0.5

    DEFAULT_TRAVEL_DIRECTION = 1 # | -1

    def __init__(self, initialValue, minBound, maxBound, cycleLog = False, DIR = 0.2):

        self.value = initialValue
        #self.ptrValue1,self.ptrValue2 = self.value,None
        self.ptrValue1 = self.value

        self.bounds = [minBound,maxBound]

        ### TODO: mod here
        self.bounds = [min(self.bounds),max(self.bounds)]
        self.DIR = DIR

        # cycle counter
        self.cycleCounter = 0
        self.cycled = False
        self.hopDirection = None
        self.initialized = False
        self.head = None
        self.headIndex = None
        self.calculate_hop_directions()
        self.elementCount = 0

        self.cycleLogActive = cycleLog
        self.cycleLog = []

        return

    def clear_pointer_values(self):
        ##self.ptrValue1,self.ptrValue2 = None,None
        self.ptrValue1 = None

    ############## start: declaring hop directions

    def head_(self):
        return self.value

    def calculate_hop_directions(self):
        self.clear_pointer_values()

        if self.value == self.bounds[0]:
            self.hopDirection = np.array([float(self.bounds[1] - self.bounds[0]) * self.DIR,\
                                self.bounds[1]])

        elif self.value == self.bounds[1]:
            self.hopDirection = np.array([float(self.bounds[0] - self.bounds[1]) * self.DIR,\
                                self.bounds[0]])

        else:
            # defaults to travelling right
            dir = self.DIR * HopPattern.DEFAULT_TRAVEL_DIRECTION
            self.hopDirection = np.array([float(self.bounds[1] - self.bounds[0]) * dir,\
                                self.bounds[1]])

        self.ptrValue1 = self.value
        self.head = self.head_()

    def set_value(self,value):
        assert value >= self.bounds[0] and value <= self.bounds[1]

        self.ptrValue1 = value

    ############## end: declaring hop directions
    def __next__(self):
        if not self.initialized:
            self.initialized = not self.initialized
            self.elementCount += 1
            if self.cycleLogActive:
                self.cycleLog = [self.value]
            return self.value
        q = self.hop_one() 
        self.cycled = self.did_cycle() 
        return q 

    def rev__next__(self):
        if not self.initialized:
            self.initialized = not self.initialized
            self.elementCount += 1
            return self.value
        return self.hop_one(True)

    def value_at(self):
        # case: single
            # l -> r
            # r -> l
        head = 0 if self.hopDirection[0] >= 0 else 1
        return HopPattern.modulo_hop(self.ptrValue1, 0.0, self.bounds, head)

    def reverse_directions(self):
        """
        calculates the reverse of hop directions
        """
        return np.array([-self.hopDirection[0], self.hopDirection[1]])

    def hop_one(self, rev = False):
        """
        hops one
        """
        q = np.copy(self.hopDirection) if not rev else self.reverse_directions()

        if self.headIndex == None:
            # make assumption based on hop direction
            head = 0 if q[0] >= 0.0 else 1
        else:
            head = self.headIndex
        q2 = HopPattern.modulo_hop(self.ptrValue1, q[0], self.bounds, head)
        self.ptrValue1 = q2
        self.elementCount += 1

        # logs value if logging is on
        if self.cycleLogActive:
            if self.did_cycle():
                self.cycleLog = [q2]
            else:
                self.cycleLog.append(q2)
        return q2

    """
    set n = 5

    to solve over-rounding errors:
        n_ = n + 2
    to solve under-rounding errors:
        n_ = n - 2

    round value to three place
    """
    @staticmethod
    def boundary_round_value(value, bounds):
        # try rounding
        v = round(value,3)

        if abs(bounds[0] - v) <= 10 ** -3: # 5
            return bounds[0]

        elif abs(bounds[1] - v) <= 10 ** -3: # 5
            return bounds[1]
        return value

    """
    'hops' the value by modulo: if hop is on
    the non-head bound,

    """
    @staticmethod
    def modulo_hop(value, hop, bounds, head):
        assert head in [0,1], "invalid head {}".format(head)

        v = value
        value = round(value + hop,5)
        value = HopPattern.boundary_round_value(value, bounds)

        # case: value at [0] or [1], endpoints,
        if value == bounds[0]:
            if head: value = bounds[1]
        elif value == bounds[1]:
            if not head: value = bounds[0]

        # case: value below [0]
        elif value < bounds[0]:
            diff = value - bounds[0]
            value = round(bounds[1] + diff,5)

        # case: value above [1]
        elif value > bounds[1]:
            diff = value - bounds[1]
            value = round(bounds[0] + diff,5)
        else:
            pass
        return round(value,5)

    ################### start: cycle checker

    def did_cycle(self):
        if not self.initialized: return self.initialized
        return True if abs(self.head - self.ptrValue1) < 10 ** -3 and self.elementCount > 1 else False

    """
    checks for cycle and updates pointer values to their modulo versions.
    """
    def cycle_check(self):
        if self.did_cycle():
            self.cycleCounter += 1
            return True
        return False

    def percentage_of_cycle_run(self):
        return abs(self.ptrValue1 - self.hopDirection[1]) / (self.bounds[1] - self.bounds[0])

    ################### end: cycle checker

'''
'''
def cycle_hop_pattern(hopPattern):
    l = []
    while not hopPattern.did_cycle():
        q2 = next(hopPattern)
        l.append(q2)

    return np.round(np.array(l), 5)

def vector_hop_in_bounds(v,h,b):
    assert is_vector(v) and is_vector(h), "invalid vectors"
    assert is_proper_bounds_vector(b), "invalid bounds"

    v2 = []
    for (i,v_) in enumerate(v):
        h_ = h[i]
        b_ = tuple(sorted(b[i]))
        v2.append(HopPattern.modulo_hop(v_,h_,b_,0))

    return np.array(v2)



from .rssi_components import *
# use for adding noise
from .numerical_generator import *
from copy import deepcopy
#: default range of relative randomness to add to a point in a relevant bound; used for mode `prg`
DEFAULT_RSSI__CR__NOISE_ADDER = np.array([[0.01,0.15]])

class ResplattingSearchSpaceIterator:
    """:class:`ResplattingSearchSpaceIterator` is a data-structure
    that relies on :class:`SearchSpaceIterator>. There are two modes for this class when it resplats on
    the original bounds:
    - "relevance zoom": requires an RChainHead to determine relevance of
    each point
    - "prg": pseudo-random generator of points in relevant bounds

    At timestamp 0, there is one relevant bound, the `bounds` parameter. 
    :class:`ResplattingSearchSpaceIterator` iterates through each bounds it considers
    relevant. When this class iterates through a bound and collects relevant points,
    a relevant sub-bound is one with a minumum relevant point and maximum relevant point
    such that the value preceding (if forward iteration) or succeeding (if backward iteration)
    the minumum relevant point is not relevant and the value succeeding (if forward iteration)
    or preceding (if backward iteration) the maximum relevant point is not relevant. 

    There is a special case that :class:`ResplattingSearchSpaceIterator` considers a sub-bound, when 
    it registers a relevant point in a bound, but the next point is not relevant, then the relevant sub-bound
    is one consisting of the relevant point and the midpoint between that and the next point.

    A generalization of the behavior of :class:`ResplattingSearchSpaceIterator` is that it narrows
    down the relevant space in an initial bound using the :class:`RChainHead` relevance function,
    and the points it generates can be exact (if "rezoom") or pseudo-random (if "prg").

    :class:`ResplattingSearchSpaceIterator` uses an :class:`RChainHead` that updates itself
    after every time it finishes iterating through a relevant bounds.The arguments it passes to
    the :class:`RChainHead` instance are: 
        
        [`self.bounds`, current bound, ssiHop] + list(`additionalUpdateArgs`)

    :param bounds: all generated points lie in this matrix
    :type bounds: np.ndarray, n x 2 matrix
    :param startPoint: n-point
    :type startPoint: np.ndarray
    :param columnOrder: order by which the columns of the reference value carry over values.
    :type columnOrder: iterable
    :param SSIHop: partition for each column of bounds
    :type SSIHop: int
    :param resplattingMode: determines how the data structure will generate new values after iterating through
                            the previous bounds and collecting relevant bounds.
    :type resplattingMode: ("relevance zoom"|"prg",:class:`RChainHead`)
    :param additionalUpdateArgs: additional arguments used to update :class:`RChainHead` when updating
        :class:`ResplattingInstructor` 
    :type additionalUpdateArgs: tuple
    """

    def __init__(self,bounds, startPoint, columnOrder = None, SSIHop = 7,\
        resplattingMode = ("relevance zoom",None), additionalUpdateArgs = (), ):
        assert is_proper_bounds_vector(bounds), "bounds "
        assert is_vector(startPoint), "invalid start point"
        assert resplattingMode[0] in {"relevance zoom", "prg"}
        assert type(resplattingMode[1]) == RChainHead, "invalid argument for resplatting mode"

        self.bounds = bounds
        self.startPoint = startPoint
        self.referencePoint = np.copy(self.startPoint)
        self.columnOrder = columnOrder
        self.SSIHop = SSIHop
        self.rm = resplattingMode

        self.iteratedOnce = False # if SSI iterated over bounds once
        self.iterateIt = False
        self.terminated = False
        #: sequence of all sub-bounds in `self.bounds` that data structure has iterated over by :class:`SearchSpaceIterator``. 
        self.rangeHistory = [np.copy(self.bounds)]

        self.declare_new_ssi(np.copy(self.bounds), np.copy(self.startPoint))

        self.ri = None
        assert type(additionalUpdateArgs) == tuple, "invalid additionalUpdateArgs"
        self.aua = additionalUpdateArgs
        self.update_resplatting_instructor()
        return

    @staticmethod
    def iterate_one_bound(rssi):

        if rssi.terminated:
            yield None
            return

        if not rssi.iteratedOnce:
            while not rssi.iteratedOnce:
                yield next(rssi)
            yield rssi.ssi.close_cycle()
            rssi.iterateIt = False

        else:
            rssi.ssi.referencePoint = rssi.ssi.de_start()

            while not rssi.iterateIt:
                yield next(rssi)
            yield rssi.ssi.close_cycle()
            rssi.iterateIt = False

    @staticmethod
    def iterate_one_batch(rssi, batchSize):
        """
        outputs points corresponding to a relevant bound

        use only with `relevance zoom`
        """
        if rssi.terminated:
            yield None
            return

        if not rssi.iteratedOnce:
            while not rssi.iteratedOnce and batchSize > 0:
                qn = next(rssi)
                if not rssi.iterateIt: yield qn
                batchSize -= 1
            if rssi.iteratedOnce:
                yield rssi.ssi.close_cycle()
                rssi.ssi.referencePoint = rssi.ssi.de_start()

                batchSize -= 1
            rssi.iterateIt = False

        while not rssi.iterateIt and batchSize > 0:
            qn = next(rssi)
            if not rssi.iterateIt: yield qn
            batchSize -= 1

        # case: bound ends before batch size output
        if rssi.iterateIt and batchSize > 0:
            yield rssi.ssi.close_cycle()
            rssi.ssi.referencePoint = rssi.ssi.de_start()

            batchSize -= 1
            rssi.iterateIt = False
            ###print("\tremainder: ", batchSize)
            return ResplattingSearchSpaceIterator.iterate_one_batch(rssi,batchSize)

    def declare_new_ssi(self,bounds, startPoint):
        '''
        Declares either a :class:`SearchSpaceIterator` or :class:`SkewedSearchSpaceIterator`
        instance for the given bounds.

        :param bounds:
        :type bounds: 
        :param startPoint:
        :type startPoint:  
        '''

        # if no specified order, default is descending
        if type(self.columnOrder) == type(None):
            self.columnOrder =ResplattingSearchSpaceIterator.column_order(bounds.shape[0],"descending")

        if is_proper_bounds_vector(bounds):
            self.ssi = SearchSpaceIterator(bounds, startPoint, self.columnOrder, self.SSIHop,cycleOn = True)
        else: # make SkewedSearchSpaceIterator
            self.ssi = SkewedSearchSpaceIterator(bounds,self.bounds,startPoint,self.columnOrder,self.SSIHop,cycleOn = True)

    def update_resplatting_instructor(self,nbs = None):
        """
        Updates :class:`ResplattingInstructor` `self.ri`. If `self.ri` is None,
        then instantiates a new :class:`ResplattingInstructor` based on the mode `self.rm\[0\]`.

        :param nbs: new bounds vector
        :return nbs: np.ndarray
        """

        if type(self.ri) == type(None):
            if self.rm[0] == "relevance zoom":
                rz = RZoom(self.rm[1])
                q = (rz,None)
            else:
                cs = CenterResplat(np.copy(self.bounds), self.rm[1], DEFAULT_RSSI__CR__NOISE_ADDER)
                q = (None,cs)

            self.ri = ResplattingInstructor(q[0],q[1])
            return False

        if self.rm[0] == "relevance zoom":

            # draw from .cache
            if type(nbs) == type(None):
                nb = self.save_rzoom_bounds_info()
                print("NEXT RANGE:")
                print(nb)

                if type(nb) == type(None): return True
                sp = np.copy(nb[:,0])
            else:# use arg<nb>
                nb = nbs[0]
                sp = nbs[1]

            if self.check_duplicate_range(nb):
                return True

            self.declare_new_ssi(nb,sp)
            # log point into range history
            self.rangeHistory.append(nb)

            # update rch here
            s = [self.bounds, nb, self.SSIHop] + list(deepcopy(self.aua))
            self.rm[1].load_update_vars(s)
            self.rm[1].update_rch()
            self.ri.rzoom = RZoom(self.rm[1])

        # TODO: optional, add update func. for png here
        return False

    def update_vars_for_rch(self): 
        return [np.copy(self.bounds),self.ssi.de_bounds(),deepcopy(self.SSIHop)] + list(deepcopy(self.aua))

    def check_duplicate_range(self,d):
        '''
        checks if range `d` is already present in `self.rangeHistory`.

        :rtype: bool
        '''
        for d_ in self.rangeHistory:
            if equal_iterables(d_,d): return True
        return False

    def save_rzoom_bounds_info(self):
        '''
        stores activation ranges of rzoom pertaining to iteration of
        the most recent relevant bounds
        '''

        # case: current rzoom a.r. not saved
        if type(self.ri.rzoom.activationRange) != type(None):
            # case: make 0-bounds
            if len(self.ri.rzoom.activationRange.shape) == 1:
                 self.ri.rzoom.activationRange = np.vstack((self.ri.rzoom.activationRange,\
                    self.ri.rzoom.activationRange)).T
            self.ri.rzoom.activationRanges.append(self.ri.rzoom.activationRange)

        self.load_activation_ranges()
        # make the next rzoom
        nb = next(self.ri)
        # case: done
        if type(nb) == type(None): return nb

        # case: fix 0-bounds
        if equal_iterables(nb[:,0],nb[:,1]):
            nb = self.fix_zero_size_activation_range(nb)
        return nb

    def load_activation_ranges(self):
        '''
        used for rm[0] == "relevance rezoom"
        '''

        additions = []
        while len(self.ri.rzoom.activationRanges) > 0:
            # pop the activation range
            ar = self.ri.rzoom.activationRanges.pop(0)

            # case: 0-size bound, modify activation range
            if equal_iterables(ar[:,0],ar[:,1]):
                ar = self.fix_zero_size_activation_range(ar)

            if type(ar) != type(None):
                self.ri.rzoomBoundsCache.append(ar)

    def fix_zero_size_activation_range(self, ar):
        '''
        fixes the zero-size activation range `ar`; new activation range is ar[0], midpoint(ar[0],next(ar[0]))
        
        :param ar: zero-size activation range
        :type ar: np.ndarray, proper bounds
        :return: a range r with r\[0\] equal to ar\[0\] that is not a zero-sized range
        :rtype: np.ndarray, proper bounds
        '''

        assert equal_iterables(ar[:,0],ar[:,1]), "not zero-size"

        # terminating condition: bounds too small
        q = self.ssi.de_bounds()
        x = np.sum(point_difference_of_improper_bounds(q,self.bounds))
        if x <= 10 ** -3:
            return None

        # save ssi location
        q = self.ssi.de_value()
        #
        self.ssi.set_value(ar[:,0])

        e1 = next(self.ssi)
        self.ssi.rev__next__()
        e2 = self.ssi.rev__next__()
        e3 = np.array([e1,e2]).T

        rv = np.ones((e3.shape[0],)) / 2.0
        p = point_on_improper_bounds_by_ratio_vector(\
            self.bounds,e3,rv)

        e3[:,1] = p
        self.ssi.set_value(q)

        x = np.sum(point_difference_of_improper_bounds(e3,self.bounds))
        if x <= 10 ** -3:
            return None

        return e3

    @staticmethod
    def column_order(k,mode = "random"):
        '''
        Outputs an ordering for k columns based on `mode`, one of
        `random`, `ascending`, `descending`.
        '''
        
        assert mode in ["random","ascending","descending"]
        s = [i for i in range(k)]
        if mode == "random":
            random.shuffle(s)
        elif mode == "descending":
            s = s[::-1]
        return np.array(s)

    def __next__(self):
        return self.get_next()

    def get_next(self):
        if self.terminated: return None
        #
        q = self.pre_next()

        # case: prg terminates
        if type(q) == type(None):
            self.terminated = True
            return None

        # log point into ResplattingInstructor
        #$%$
        self.ri.output(q)

        # set a new ssi based on resplatting mode
        self.post_next()
        #if self.iterateIt: return None
        return q

    def pre_next(self):
        # case: RSSI switched to prg
        if self.iteratedOnce and self.rm[0] == "prg":
            return next(self.ri)
        return next(self.ssi)

    def post_next(self):
        if self.ssi.reached_end():
            self.iteratedOnce = True
            self.iterateIt = True
            self.terminated = self.update_resplatting_instructor()
        else:
            self.iterateIt = False

    def display_range_history(self):
        for (i,r) in enumerate(self.rangeHistory):
            print("{}:\t{}".format(i,r))

    def _summary(self):
        print("BOUND SUMMARY")
        print("parent bounds")
        print(self.bounds)
        print()
        print("start point")
        print(self.startPoint)
        print()
        print("reference point")
        print(self.referencePoint)
        print()

        #### do range history
        print("range history")
        self.display_range_history()
        #### do

def rssi__display_n_bounds(rssi, n):
    for i in range(n):
        print("iterating bound ",i)
        q = ResplattingSearchSpaceIterator.iterate_one_bound(rssi)
        for q_ in q:
            print(q_)
        # summary
        rssi._summary()
        print("\n--------------------------------------------------")
    return -1
